"""
========================================================================
ComponentLevel3.py
========================================================================
We add value wire/interface connections. Basically, all connected
value signal in the whole graph should have the same value of the unique
"net writer" written in an update block.
Then, the update block for a net is basically one writer writes to those
readers. Interface connections are handled separately, and they should
be revamped when adding method-based interfaces.

Author : Shunning Jiang
Date   : Jan 29, 2020
"""
import ast
import inspect
import linecache
from collections import defaultdict

from pymtl3.datatypes import Bits, is_bitstruct_inst

from .ComponentLevel1 import ComponentLevel1
from .ComponentLevel2 import ComponentLevel2, compiled_re
from .Connectable import (
    Connectable,
    Const,
    InPort,
    Interface,
    OutPort,
    Signal,
    Wire,
    _connect_check,
)
from .errors import (
    InvalidConnectionError,
    InvalidPlaceholderError,
    MultiWriterError,
    NotElaboratedError,
    NoWriterError,
    PyMTLDeprecationError,
    SignalTypeError,
)
from .NamedObject import NamedObject
from .Placeholder import Placeholder


def connect( o1, o2 ):
  host, o1_connectable, o2_connectable = _connect_check( o1, o2, internal=False )
  host._connect_dispatch( o1, o2, o1_connectable, o2_connectable )

class ComponentLevel3( ComponentLevel2 ):

  #-----------------------------------------------------------------------
  # Private methods
  #-----------------------------------------------------------------------

  def __new__( cls, *args, **kwargs ):
    inst = super().__new__( cls, *args, **kwargs )
    inst._dsl.call_kwargs   = None
    inst._dsl.adjacency     = defaultdict(set)
    inst._dsl.connect_order = []
    inst._dsl.consts        = set()

    return inst

  # Override
  def _collect_vars( s, m ):
    super()._collect_vars( m )
    if isinstance( m, ComponentLevel3 ):
      all_ajd = s._dsl.all_adjacency
      for k, v in m._dsl.adjacency.items():
        all_ajd[k] |= v

  # Override
  def _construct( s ):
    """ We override _construct here to finish the saved __call__
    connections right after constructing the model. The reason why we
    take this detour instead of connecting in __call__ directly, is that
    __call__ is done before setattr, and hence the child components don't
    know their name yet. _dsl.constructed is called in setattr after name
    tagging, so this is valid. (see NamedObject.py). """

    if not s._dsl.constructed:

      # Merge the actual keyword args and those args set by set_parameter
      if s._dsl.param_tree is None:
        kwargs = s._dsl.kwargs
      elif s._dsl.param_tree.leaf is None:
        kwargs = s._dsl.kwargs
      else:
        kwargs = s._dsl.kwargs.copy()
        if "construct" in s._dsl.param_tree.leaf:
          more_args = s._dsl.param_tree.leaf[ "construct" ]
          kwargs.update( more_args )

      s.construct( *s._dsl.args, **kwargs )

      if s._dsl.call_kwargs is not None: # s.a = A()( b = s.b )
        s._continue_call_connect()

      s._dsl.constructed = True

  # The following three methods should only be called when types are
  # already checked
  def _create_assign_lambda( s, o, lamb ):
    assert isinstance( o, Signal ), "You can only assign(//=) a lambda function to a Wire/InPort/OutPort."

    srcs, line = inspect.getsourcelines( lamb )
    assert len(srcs) == 1, "We can only handle single-line lambda connect right now."

    src  = compiled_re.sub( r'\2', srcs[0] ).lstrip(' ')
    root = ast.parse(src)
    assert isinstance( root, ast.Module ) and len(root.body) == 1, "Invalid lambda (contact pymtl3 developer)"

    root = root.body[0]
    assert isinstance( root, ast.AugAssign ) and isinstance( root.op, ast.FloorDiv )

    lhs, rhs = root.target, root.value
    # We expect the lambda to have no argument:
    # {'args': [], 'vararg': None, 'kwonlyargs': [], 'kw_defaults': [], 'kwarg': None, 'defaults': []}
    assert isinstance( rhs, ast.Lambda ) and not rhs.args.args and rhs.args.vararg is None, \
      "The lambda shouldn't contain any argument."

    rhs = rhs.body

    # Compose a new and valid function based on the lambda's lhs and rhs
    # Note that we don't need to add those source code of closure var
    # assignment to linecache. To get the matching line number in the
    # error message, we set the line number of update block
    # Shunning: bugfix:

    blk_name = "_lambda__{}".format( repr(o).replace(".","_").replace("[", "_").replace("]", "_") )
    lambda_upblk = ast.FunctionDef(
      name=blk_name,
      args=ast.arguments(args=[], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]),
      body=[ast.Assign(targets=[lhs], value=rhs, lineno=2, col_offset=6)],
      decorator_list=[],
      returns=None,
      lineno=1, col_offset=4,
    )
    lambda_upblk_module = ast.Module(body=[ lambda_upblk ])

    # Manually wrap the lambda upblk with a closure function that adds the
    # desired variables to the closure of `_lambda__*`
    # We construct AST for the following function to add free variables in the
    # closure of the lambda function to the closure of the generated lambda
    # update block.
    #
    # def closure( lambda_closure ):
    #   <FreeVarName1> = lambda_closure[<Idx1>].cell_contents
    #   <FreeVarName2> = lambda_closure[<Idx2>].cell_contents
    #   ...
    #   <FreeVarNameN> = lambda_closure[<IdxN>].cell_contents
    #   def _lambda__<lambda_blk_name>():
    #     # the assignment statement appears here
    #   return _lambda__<lambda_blk_name>

    new_root = ast.Module( body=[
      ast.FunctionDef(
          name="closure",
          args=ast.arguments(args=[ast.arg(arg="lambda_closure", annotation=None, lineno=1, col_offset=12)],
                             vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]),
          body=[
            ast.Assign(
              targets=[ast.Name(id=var, ctx=ast.Store(), lineno=1+idx, col_offset=2)],
              value=ast.Attribute(
                value=ast.Subscript(
                  value=ast.Name(
                    id='lambda_closure',
                    ctx=ast.Load(),
                    lineno=1+idx, col_offset=5+len(var),
                  ),
                  slice=ast.Index(
                    value=ast.Num(
                      n=idx,
                      lineno=1+idx, col_offset=19+len(var),
                    ),
                  ),
                  ctx=ast.Load(),
                  lineno=1+idx, col_offset=5+len(var),
                ),
                attr='cell_contents',
                ctx=ast.Load(),
                lineno=1+idx, col_offset=5+len(var),
              ),
              lineno=1+idx, col_offset=2,
            ) for idx, var in enumerate(lamb.__code__.co_freevars)
          ] + [ lambda_upblk ] + [
            ast.Return(
              value=ast.Name(
                id=blk_name,
                ctx=ast.Load(),
                lineno=4+len(lamb.__code__.co_freevars), col_offset=9,
              ),
              lineno=4+len(lamb.__code__.co_freevars), col_offset=2,
            )
          ],
          decorator_list=[],
          returns=None,
          lineno=1, col_offset=0,
        )
    ] )

    # In Python 3 we need to supply a dict as local to get the newly
    # compiled function from closure.
    # Then `closure(lamb.__closure__)` returns the lambda update block with
    # the correct free variables in its closure.

    dict_local = {}
    exec( compile(new_root, blk_name, "exec"), lamb.__globals__, dict_local )
    blk = dict_local[ 'closure' ]( lamb.__closure__ )

    # Add the source code to linecache for the compiled function

    new_src = "def {}():\n {}\n".format( blk_name, src.replace("//=", "=") )
    linecache.cache[ blk_name ] = (len(new_src), None, new_src.splitlines(), blk_name)

    ComponentLevel1.update( s, blk )

    # This caching here does no caching because the block name contains
    # the signal name intentionally to avoid conflicts. With //= it is
    # more possible than normal update block to have conflicts:
    # if param == 1:  s.out //= s.in_ + 1
    # else:           s.out //= s.out + 100
    # Here these two blocks will implicity have the same name but they
    # have different contents based on different param.
    # So the cache call here is just to reuse the existing interface to
    # register the AST/src of the generated block for elaborate or passes
    # to use.
    s._cache_func_meta( blk, is_update_ff=False,
      given=("".join(srcs), lambda_upblk_module, line, inspect.getsourcefile( lamb )) )
    return blk

  def _connect_signal_const( s, o1, o2 ):
    Type = o1._dsl.Type
    if isinstance( o2, int ):
      if not issubclass( Type, (int, Bits) ):
        raise InvalidConnectionError( f"We don't support connecting an integer constant "
                                      f"to non-int/Bits type {Type}" )
      o2 = Const( Type, Type(o2), s )
    elif isinstance( o2, Bits ):
      Type2 = type(o2)
      if not issubclass( Type, Bits ):
        raise InvalidConnectionError( f"We don't support connecting a {Type2} constant "
                                      f"to non-Bits type {Type}" )
      if Type is not Type2:
        raise InvalidConnectionError( f"Bitwidth mismatch when connecting a {Type2} constant "
                                      f"to signal {o1} with type {Type}." )
      o2 = Const( Type, o2, s )
    elif is_bitstruct_inst( o2 ):
      Type2 = type(o2)
      if Type is not Type2:
        raise InvalidConnectionError( f"We don't support connecting a {Type2} constant bitstruct"
                                      f"to non-bitstruct type {Type}" )
      o2 = Const( Type, o2, s )
    else:
      raise InvalidConnectionError(f"\n>>> {o2} of type {type(o2)} is not a const! \n"
                                   f">>> It cannot be connected to signal {o1} of type {o1._dsl.Type}!\n"
                                   f"Suggestion: fix the RHS of connection to be an instance of {o1._dsl.Type}.")

    # TODO implement connecting a const struct

    host = o1.get_host_component()

    if isinstance( o1, InPort ):
      # connecting constant to inport should be at the parent level
      host = host.get_parent_object()

    o2._dsl.parent_obj = s
    s._dsl.consts.add( o2 )

    # o2 should be a new object

    s._dsl.adjacency[o1].add( o2 )
    s._dsl.adjacency[o2].add( o1 )

    s._dsl.connect_order.append( (o1, o2) )

  def _connect_signal_signal( s, o1, o2 ):
    if not (o1._dsl.Type is o2._dsl.Type):
      raise InvalidConnectionError( f"Type mismatch {o1._dsl.Type} != {o2._dsl.Type} during {o1}<->{o2}." )

    if o1 not in s._dsl.adjacency[o2]:
      assert o2 not in s._dsl.adjacency[o1]
      s._dsl.adjacency[o1].add( o2 )
      s._dsl.adjacency[o2].add( o1 )

      s._dsl.connect_order.append( (o1, o2) )

  def _connect_interfaces( s, o1, o2 ):
    # When we connect two interfaces, we first try to use o1's and o2's
    # connect. If failed, we fall back to by-name connection

    def connect_by_name( this, other ):
      def recursive_connect( this_obj, other_obj ):
        if isinstance( this_obj, list ):
          for i in range(len(this_obj)):
            # TODO add error message if other_obj is not a list
            recursive_connect( this_obj[i], other_obj[i] )
        else:
          s._connect( other_obj, this_obj, internal=True )

      # Sort the keys to always connect in a unique order
      for name in sorted(this.__dict__):
        if name[0] != '_': # filter private variables
          obj = this.__dict__[ name ]
          if hasattr( other, name ):
            # other has the corresponding field, connect recursively
            recursive_connect( obj, getattr( other, name ) )

          else:
            # other doesn't have the corresponding field, raise error
            # if obj is connectable.
            if isinstance( obj, Connectable ):
              raise InvalidConnectionError("There is no \"{}\" field in {} "
              "to connect to {} during by-name connection\n"
              "Suggestion: check the implementation of \n"
              "          - {} (class {})\n"
              "          - {} (class {})".format( name, other, obj,
                repr(this), type(this), repr(other), type(other) ) )

    if hasattr( o1, "connect" ):
      if not o1.connect( o2, s ): # o1.connect fail
        if hasattr( o2, "connect" ):
          if not o2.connect( o1, s ):
            connect_by_name( o1, o2 )
        else:
          connect_by_name( o1, o2 )

    else: # o1 has no "connect"
      if hasattr( o2, "connect" ):
        if not o2.connect( o1, s ):
          connect_by_name( o1, o2 )
      else:
        connect_by_name( o1, o2 ) # capture s

  def _connect( s, o1, o2, internal ):
    """ Top level private method for connecting two objects. """
    host, o1_connectable, o2_connectable = _connect_check( o1, o2, internal )

    if host is None:
      # host is None only happens when internal = True
      assert internal

    else:
      assert s is host, "Please contact pymtl3 developer -- s:{} host:{}".format(s, host)
      s._connect_dispatch( o1, o2, o1_connectable, o2_connectable )

  def _connect_dispatch( s, o1, o2, o1_connectable, o2_connectable ):

    if o1_connectable and o2_connectable:
      # if both connectable, dispatch signal-signal and interface-interface
      if isinstance( o1, Signal ) and isinstance( o2, Signal ):
        s._connect_signal_signal( o1, o2 )
      elif isinstance( o1, Interface ) and isinstance( o2, Interface ):
        s._connect_interfaces( o1, o2 )
      else:
        raise InvalidConnectionError("{} cannot be connected to {}: {} != {}" \
              .format(repr(o1), repr(o2), type(o1), type(o2)) )
    else:
      # One is connectable, we make sure it's o1
      if o2_connectable:
        o1, o2 = o2, o1
      assert isinstance( o1, Signal ), f"Cannot connect {o1!r} to {o2!r}."

      s._connect_signal_const( o1, o2 )

  def _continue_call_connect( s ):
    """ Here we continue to establish the connections from signals of the
    parent object, to signals in the current object. Since it is the
    parent that connects a constant integer to a signal, we should point
    the Const object back to the parent object by setting _parent_obj to
    s._parent_obj."""

    parent = s._dsl.parent_obj

    top = s._dsl.elaborate_top

    # _continue_call_connect is actually connecting stuff at parent level,
    # but it currently happens before before we pop the child component.
    # To make minimal modification, I temporarily pop it from
    # elaborate_stack and append it back at the end of this method

    tmp = top._dsl.elaborate_stack.pop()

    try: # Catch AssertionError from _connect

      # Process saved __call__ kwargs
      for (kw, target) in s._dsl.call_kwargs.items():
        try:
          obj = getattr( s, kw )
        except AttributeError:
          raise InvalidConnectionError( "{} is not a member of class {}".format(kw, s.__class__) )

        # Obj is a list of signals
        # We assume the a dict of { index: obj } is provided.
        if   isinstance( obj, list ):
          # Make sure the connection target is a dictionary {idx: obj}
          if not isinstance( target, dict ):
            raise InvalidConnectionError( "We only support a dictionary when '{}' is an array.".format( kw ) )
          for idx, item in target.items():
            parent._connect( obj[idx], item, internal=False )

        # Obj is a single signal
        # If the target is a list, it's fanout connection
        elif isinstance( target, (tuple, list) ):
          for item in target:
            parent._connect( obj, item, internal=False )

        # Target is a single object
        else:
          parent._connect( obj, target, internal=False )

    except AssertionError as e:
      raise InvalidConnectionError( "Invalid connection for {}:\n{}".format( kw, e ) )

    # Append tmp back to elaborate_stack
    top._dsl.elaborate_stack.append(tmp)

  @staticmethod
  def _floodfill_nets( signal_list, adjacency ):
    """ Floodfill to find out connected nets. Return a list of sets. """

    nets = []
    visited = set()
    pred    = {} # detect cycle that has >=3 nodes
    for obj in signal_list:
      # If obj has adjacent signals
      if obj in adjacency and obj not in visited:
        net = set()
        Q   = [ obj ]
        while Q:
          u = Q.pop()
          visited.add( u )
          net.add( u )
          for v in adjacency[u]:
            if v not in visited:
              pred[v] = u
              Q.append( v )
            elif v is not pred[u]:
              raise InvalidConnectionError(repr(v)+" is in a connection loop.")
        if len(net) == 1:
          continue
        nets.append( net )
    return nets

  def _resolve_value_connections( s ):
    """ The case of nested data struct: the writer of a net can be one of
    the three: signal itself (s.x.a), ancestor (s.x), descendant (s.x.b)

    An iterative algorithm is required to mark the writers. The example
    is the following. Net 1's writer is s.x and one reader is s.y.
    Net 2's writer is s.y.a (known ONLY after Net 1's writer is clear),
    one reader is s.z. Net 3's writer is s.z.a (known ...), and so forth

    Note that s.x becomes writer when WR s.x.a or WR s.x.b, but s.x then
    cannot propagate back to s.x.b or s.x.a.

    The original state is all the writers from all update blocks.
    writer_prop is a dict {x:y} that stores potential writers and
    whether the writer can propagate to other nets. After a net is
    resolved from headless condition, its readers become writers.

    The case of slicing: slices of the same wire are only one level
    deeper, so all of those parent/child relationship work easily.
    However, unlike different fields of a data struct, different slices
    may _intersect_, so they need to check sibling slices' write/read
    status as well. """

    # First of all, bfs the "forest" to find out all nets

    nets = s._floodfill_nets( s._dsl.all_signals, s._dsl.all_adjacency )

    # Then figure out writers: all writes in upblks and their nest objects

    writer_prop = {}

    for blk, writes in s._dsl.all_upblk_writes.items():
      for obj in writes:
        writer_prop[ obj ] = True # propagatable

        obj = obj.get_parent_object()
        while obj.is_signal():
          writer_prop[ obj ] = False
          obj = obj.get_parent_object()

    # Find the host object of every net signal
    # and then leverage the information to find out top level input port

    for net in nets:
      for member in net:
        host = member.get_host_component()

        # Specialize two cases:
        # 1. A top-level input port is writer.
        # 2. An output port of a placeholder module is a writer
        if ( isinstance( member, InPort ) and host == s ) or \
           ( isinstance( member, OutPort ) and isinstance( host, Placeholder ) ):
          writer_prop[ member ] = True

    headless = nets
    headed   = []

    # Convention: we store a net in a tuple ( writer, set([readers]) )
    # The first element is writer; it should be None if there is no
    # writer. The second element is a set of signals including the writer.

    while headless:
      new_headless = []
      wcount = len(writer_prop)

      # For each net, figure out the writer among all vars and their
      # ancestors. Moreover, if x's ancestor has a writer in another net,
      # x should be the writer of this net.
      #
      # If there is a writer, propagate writer information to all readers
      # and readers' ancestors. The propagation is tricky: assume s.x.a
      # is in net, and s.x.b is written in upblk, s.x.b will mark s.x as
      # an unpropagatable writer because later s.x.a shouldn't be marked
      # as writer by s.x.
      #
      # Similarly, if x[0:10] is written in update block, x[5:15] can
      # be a unpropagatable writer because we don't want x[5:15] to
      # propagate to x[12:17] later.

      for net in headless:
        has_writer = False

        for v in net:
          obj = None
          try:
            # Check if itself is a writer or a constant
            if v in writer_prop or isinstance( v, Const ):
              assert not has_writer
              has_writer, writer = True, v

            else:
              # Check if an ancestor is a propagatable writer
              obj = v.get_parent_object()
              while obj.is_signal():
                if obj in writer_prop and writer_prop[ obj ]:
                  assert not has_writer
                  has_writer, writer = True, v
                  break
                obj = obj.get_parent_object()

              # Check sibling slices
              for obj in v.get_sibling_slices():
                if obj.slice_overlap( v ):
                  if obj in writer_prop and writer_prop[ obj ]:
                    assert not has_writer
                    has_writer, writer = True, v
                    # Shunning: is breaking out of here enough? If we
                    # don't break the loop, we might a list here storing
                    # "why the writer became writer" and do some sibling
                    # overlap checks when we enter the loop body later
                    break

          except AssertionError:
            raise MultiWriterError( \
            "Two-writer conflict \"{}\"{}, \"{}\" in the following net:\n - {}".format(
              repr(v), "" if not obj else "(as \"{}\" is written somewhere else)".format( repr(obj) ),
              repr(writer), "\n - ".join([repr(x) for x in net])) )

        if not has_writer:
          new_headless.append( net )
          continue

        # Child s.x.y of some propagatable s.x, or sibling of some
        # propagatable s[a:b].
        # This means that at least other variables are able to see s.x/s[a:b]
        # so it doesn't matter if s.x.y is not in writer_prop
        if writer not in writer_prop:
          pass

        for v in net:
          if v != writer:
            writer_prop[ v ] = True # The reader becomes new writer

            obj = v.get_parent_object()
            while obj.is_signal():
              if obj not in writer_prop:
                writer_prop[ obj ] = False
              obj = obj.get_parent_object()

        headed.append( (writer, net) )

      if wcount == len(writer_prop): # no more new writers
        break
      headless = new_headless

    return headed + [ (None, x) for x in headless ]

  def _check_port_in_nets( s ):
    nets = s._dsl.all_value_nets

    # The case of connection is very tricky because we put a single upblk
    # in the lowest common ancestor node and the "output port" chain is
    # inverted. So we need to deal with it here ...
    #
    # The gist is that the data flows from deeper level writer to upper
    # level readers via output port, to the same level via wire, and from
    # upper level to deeper level via input port

    headless = [ signals for writer, signals in nets if writer is None ] # remove None
    if headless:
      raise NoWriterError( headless )

    for writer, _ in nets:

      # We need to do DFS to check all connected port types
      # Each node is a writer when we expand it to other nodes

      S = [ writer ]
      visited = { writer }

      while S:
        u = S.pop() # u is the writer
        whost = u.get_host_component()

        for v in s._dsl.all_adjacency[u]: # v is the reader
          if v not in visited:
            visited.add( v )
            S.append( v )
            rhost = v.get_host_component()

            # 1. have the same host: writer_host(x)/reader_host(x):
            # Hence, writer is anything, reader is wire or outport
            if   whost == rhost:
              valid = isinstance( u, (Signal, Const) ) and \
                      isinstance( v, (OutPort, Wire) )
              if not valid:
                # Check if it's an outport driving inport. If it is
                # connected at parent level, we permit this loopback from
                # parent level.
                if isinstance( u, OutPort ) and isinstance( v, InPort ):
                  u_connected_in_whost = v in whost._dsl.adjacency and u in whost._dsl.adjacency[v]
                  v_connected_in_whost = u in whost._dsl.adjacency and v in whost._dsl.adjacency[u]
                  assert u_connected_in_whost == v_connected_in_whost, "Please contact pymtl3 developers."

                  parent = whost.get_parent_object()
                  u_connected_in_parent = v in parent._dsl.adjacency and u in parent._dsl.adjacency[v]
                  v_connected_in_parent = u in parent._dsl.adjacency and v in parent._dsl.adjacency[u]
                  assert u_connected_in_parent == v_connected_in_parent, "Please contact pymtl3 developers."

                  assert u_connected_in_whost != u_connected_in_parent, "Please contact pymtl3 developers."

                  # We permit this loopback from parent level. Otherwise
                  # we throw an error
                  if not u_connected_in_parent:
                    raise InvalidConnectionError( \
"""InPort and OutPort loopback connection is only allowed at parent level:

- Unless the connection is fulfilled in parent "{}",
  {} "{}" of {} (class {}) cannot be driven by {} "{}" of {} (class {}).

  Note: Looks like the connection is fulfilled in "{}".""" \
          .format(  parent,
                    type(v).__name__, repr(v), repr(rhost), type(rhost).__name__,
                    type(u).__name__, repr(u), repr(whost), type(whost).__name__,
                    repr(whost) ) )

                else:
                  raise SignalTypeError( \
"""[Type 5] Invalid port type detected at the same host component "{}" (class {})

- {} "{}" cannot be driven by {} "{}".

  Note: InPort x.y cannot be driven by x.z""" \
          .format(  repr(rhost), type(rhost).__name__,
                    type(v).__name__, repr(v), type(u).__name__, repr(u) ) )

            # 2. reader_host(x) is writer_host(x.y)'s parent:
            # Hence, writer is outport, reader is wire or outport
            # writer cannot be constant
            elif rhost == whost.get_parent_object():
              valid = isinstance( u, OutPort ) and \
                      isinstance( v, (OutPort, Wire) )

              if not valid:
                raise SignalTypeError( \
"""[Type 6] Invalid port type detected when the driver lies deeper than drivee:

- {} "{}" of {} (class {}) cannot be driven by {} "{}" of {} (class {}).

  Note: InPort x.y cannot be driven by x.z.a""" \
          .format(  type(v).__name__, repr(v), repr(rhost), type(rhost).__name__,
                    type(u).__name__, repr(u), repr(whost), type(whost).__name__ ) )

            # 3. writer_host(x) is reader_host(x.y)'s parent:
            # Hence, writer is inport or wire, reader is inport
            # writer can be constant
            elif whost == rhost.get_parent_object():
              # valid = ( isinstance( u, InPort ) or isinstance( u, Wire ) \
                                                 # or isinstance( u, Const)) and \
                         # isinstance( v, InPort )

              # if not valid:
                # raise SignalTypeError( \
# """[Type 7] Invalid port type detected when the driver lies shallower than drivee:

# - {} "{}" of {} (class {}) cannot be driven by {} "{}" of {} (class {}).

  # Note: OutPort/Wire x.y.z cannot be driven by x.a""" \
          # .format(  type(v).__name__, repr(v), repr(rhost), type(rhost).__name__,
                    # type(u).__name__, repr(u), repr(whost), type(whost).__name__ ) )

            # Shunning 9/12/2017: Actually in this case writer can be outport
              valid = isinstance( u, (Signal, Const) ) and isinstance( v, InPort )

              if not valid:
                raise SignalTypeError( \
"""[Type 7] Invalid port type detected when the driver lies shallower than drivee:

- {} "{}" of {} (class {}) cannot be driven by {} "{}" of {} (class {}).

  Note: OutPort/Wire x.y.z cannot be driven by x.a""" \
          .format(  type(v).__name__, repr(v), repr(rhost), type(rhost).__name__,
                    type(u).__name__, repr(u), repr(whost), type(whost).__name__ ) )

            # 4. hosts have the same parent: writer_host(x.y)/reader_host(x.z)
            # This means that the connection is fulfilled in x
            # Hence, writer is outport and reader is inport
            # writer cannot be constant
            elif whost.get_parent_object() == rhost.get_parent_object():
              valid = isinstance( u, OutPort ) and isinstance( v, InPort )

              if not valid:
                raise SignalTypeError( \
"""[Type 8] Invalid port type detected when the drivers is the sibling of drivee:

- {} "{}" of {} (class {}) cannot be driven by {} "{}" of {} (class {}).

  Note: Looks like the connection is fulfilled in "{}".
        OutPort/Wire x.y.z cannot be driven by x.a.b""" \
          .format(  type(v).__name__, repr(v), repr(rhost), type(rhost).__name__,
                    type(u).__name__, repr(u), repr(whost), type(whost).__name__,
                    repr(whost.get_parent_object()) ) )
            # 5. neither host is the other's parent nor the same.
            else:
              raise SignalTypeError("""[Type 9] "{}" and "{}" cannot be connected:

- host objects "{}" and "{}" are too far in the hierarchy.""" \
              .format( repr(u), repr(v), repr(whost), repr(rhost) ) )

  def _disconnect_signal_int( s, o1, o2 ):

    nets = s.get_all_value_nets()

    for i, net in enumerate( nets ):
      writer, signals = net

      if o1 in signals: # Find the net that contains o1
        # If we're disconnecting a constant from a port, the constant
        # should be the only writer in this net and is equal to o2
        assert isinstance( writer, Const ), "what the hell?"
        assert writer._dsl.const == o2, "Disconnecting the wrong const {} " \
                                        "-- should be {}.".format( o2, writer.const )
        o2 = writer

        # I don't remove it from m._adjacency since they are not used later
        assert o1 in s._dsl.all_adjacency[o2] and o2 in s._dsl.all_adjacency[o1]
        s._dsl.all_adjacency[o2].remove( o1 )
        s._dsl.all_adjacency[o1].remove( o2 )

        # Disconnect a const from a signal just removes the writer in the net
        signals.remove( writer )
        nets[i] = ( None, signals )
        return

  def _disconnect_signal_signal( s, o1, o2 ):

    nets = s.get_all_value_nets()

    assert o1 in s._dsl.all_adjacency[o2] and o2 in s._dsl.all_adjacency[o1]
    # I don't remove it from m._adjacency since they are not used later
    s._dsl.all_adjacency[o2].remove( o1 )
    s._dsl.all_adjacency[o1].remove( o2 )

    for i, net in enumerate( nets ):
      writer, signals = net

      if o1 in signals: # Find the net that contains o1
        assert o2 in signals, signals

        broken_nets = s._floodfill_nets( signals, s._dsl.all_adjacency )

        # disconnect the only two vertices in the net
        if len(broken_nets) == 0:
          nets[i] = nets.pop() # squeeze in the last net

        # the removed edge results in an isolated vertex and a connected component
        elif len(broken_nets) == 1:
          net0 = broken_nets[0]
          if writer in net0:
            nets[i] = ( writer, net0 )
          else:
            assert writer is o1 or writer is o2
            nets[i] = ( None, net0 )

        elif len(broken_nets) == 2:
          net0, net1 = broken_nets[0], broken_nets[1]
          if writer in net0:
            nets[i] = ( writer, net0 ) # replace in-place
            nets.append( (None, net1) )
          else:
            assert writer in net1
            nets[i] = ( None, net0 ) # replace in-place
            nets.append( (writer, net1) )

        else:
          assert False, "what the hell?"

        return

  # Override
  def _check_valid_dsl_code( s ):
    s._check_upblk_writes()
    s._check_port_in_upblk()
    s._check_port_in_nets()

  #-----------------------------------------------------------------------
  # Construction-time APIs
  #-----------------------------------------------------------------------

  def __call__( s, *args, **kwargs ):
    """ This syntactic sugar supports the following one-liner:
      >>> s.x = SomeReg(Bits1)( in_ = s.in_ )
    It connects s.in_ to s.x.in_ in the same line as model construction.
    """
    assert args == ()
    if s._dsl.constructed:
      raise InvalidConnectionError("Connection using __call__, "
                                   "i.e. s.x( a = s.a ), is illegal "
                                   "after constructing s.x")
    s._dsl.call_kwargs = kwargs
    return s

  def connect( s, *args, **kwargs ):
    raise PyMTLDeprecationError("\ns.connect method has been deprecated! "
                                "\n- Please use free function connect(s.x,s.y) or "
                                "syntactic sugar s.x//=s.y instead.")

  #-----------------------------------------------------------------------
  # elaborate
  #-----------------------------------------------------------------------
  # Since the spawned signals are handled by the updated elaborate
  # template in ComponentLevel2, we just need to add a bit more
  # functionalities to handle nets.

  # Override
  def _elaborate_declare_vars( s ):
    super()._elaborate_declare_vars()
    s._dsl.all_adjacency = defaultdict(set)

  # Override
  def _elaborate_collect_all_vars( s ):
    super()._elaborate_collect_all_vars()
    s._dsl.all_value_nets = s._resolve_value_connections()
    s._dsl._has_pending_value_connections = False

    s._check_valid_dsl_code()

  #-----------------------------------------------------------------------
  # Post-elaborate public APIs (can only be called after elaboration)
  #-----------------------------------------------------------------------
  # We have moved these implementations to Component.py because the
  # outside world should only use Component.py

  def get_all_value_nets( s ):

    if s._dsl._has_pending_value_connections:
      s._dsl.all_value_nets = s._resolve_value_connections()
      s._dsl._has_pending_value_connections = False

    return s._dsl.all_value_nets



#=========================================================================
# TestMemory_test.py
#=========================================================================

import random
import struct

import pytest

from pymtl3 import *
from pymtl3.stdlib.ifcs import MemMsgType, mk_mem_msg
from pymtl3.stdlib.test import mk_test_case_table
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL

from .MemoryCL import MemoryCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, cls, nports, PortTypes, src_msgs, sink_msgs,
                 stall_prob, mem_latency, src_initial,  src_interval, sink_initial, sink_interval,
                 arrival_time=None ):
    assert len(PortTypes) == nports
    s.srcs = [ TestSrcCL( PortTypes[i][0], src_msgs[i], src_initial, src_interval )
                for i in range(nports) ]
    s.mem  = cls( nports, PortTypes, mem_latency )
    s.sinks = [ TestSinkCL( PortTypes[i][1], sink_msgs[i], sink_initial, sink_interval,
                            arrival_time ) for i in range(nports) ]

    # Connections
    for i in range(nports):
      connect( s.srcs[i].send, s.mem.ifc[i].req )
      connect( s.mem.ifc[i].resp,  s.sinks[i].recv  )

  def done( s ):
    return all([x.done() for x in s.srcs] + [x.done() for x in s.sinks])

  def line_trace( s ):
    return "{} >>>  {}  >>> {}".format(
      "|".join( [ x.line_trace() for x in s.srcs ] ),
      s.mem.line_trace(),
      "|".join( [ x.line_trace() for x in s.sinks ] ) )

#-------------------------------------------------------------------------
# make messages
#-------------------------------------------------------------------------

req_type_dict = {
  'rd': MemMsgType.READ,
  'wr': MemMsgType.WRITE,
  'ad': MemMsgType.AMO_ADD,
  'an': MemMsgType.AMO_AND,
  'or': MemMsgType.AMO_OR,
  'xg': MemMsgType.AMO_SWAP,
  'mn': MemMsgType.AMO_MIN,
}

resp_type_dict = {
  'rd': MemMsgType.READ,
  'wr': MemMsgType.WRITE,
  'ad': MemMsgType.AMO_ADD,
  'an': MemMsgType.AMO_AND,
  'or': MemMsgType.AMO_OR,
  'xg': MemMsgType.AMO_SWAP,
  'mn': MemMsgType.AMO_MIN,
}

req_cls, resp_cls = mk_mem_msg( 8, 32, 32 )
b32 = Bits32

def req( type_, opaque, addr, len, data ):
  return req_cls( req_type_dict[type_], opaque, addr, len, b32(data) )

def resp( type_, opaque, len, data ):
  return resp_cls( resp_type_dict[type_], opaque, 0, len, b32(data) )

#----------------------------------------------------------------------
# Test Case: basic
#----------------------------------------------------------------------

def basic_msgs( base_addr ):
  return [
    req( 'wr', 0x0, base_addr, 0, 0xdeadbeef ), resp( 'wr', 0x0, 0, 0          ),
    req( 'rd', 0x1, base_addr, 0, 0          ), resp( 'rd', 0x1, 0, 0xdeadbeef ),
  ]

#----------------------------------------------------------------------
# Test Case: stream
#----------------------------------------------------------------------

def stream_msgs( base_addr ):

  msgs = []
  for i in range(20):
    msgs.extend([
      req( 'wr', i, base_addr+4*i, 0, i ), resp( 'wr', i, 0, 0 ),
      req( 'rd', i, base_addr+4*i, 0, 0 ), resp( 'rd', i, 0, i ),
    ])

  return msgs

#----------------------------------------------------------------------
# Test Case: subword reads
#----------------------------------------------------------------------

def subword_rd_msgs( base_addr ):
  return [

    req( 'wr', 0x0, base_addr+0, 0, 0xdeadbeef ), resp( 'wr', 0x0, 0, 0          ),

    req( 'rd', 0x1, base_addr+0, 1, 0          ), resp( 'rd', 0x1, 1, 0x000000ef ),
    req( 'rd', 0x2, base_addr+1, 1, 0          ), resp( 'rd', 0x2, 1, 0x000000be ),
    req( 'rd', 0x3, base_addr+2, 1, 0          ), resp( 'rd', 0x3, 1, 0x000000ad ),
    req( 'rd', 0x4, base_addr+3, 1, 0          ), resp( 'rd', 0x4, 1, 0x000000de ),

    req( 'rd', 0x5, base_addr+0, 2, 0          ), resp( 'rd', 0x5, 2, 0x0000beef ),
    req( 'rd', 0x6, base_addr+1, 2, 0          ), resp( 'rd', 0x6, 2, 0x0000adbe ),
    req( 'rd', 0x7, base_addr+2, 2, 0          ), resp( 'rd', 0x7, 2, 0x0000dead ),

    req( 'rd', 0x8, base_addr+0, 3, 0          ), resp( 'rd', 0x8, 3, 0x00adbeef ),
    req( 'rd', 0x9, base_addr+1, 3, 0          ), resp( 'rd', 0x9, 3, 0x00deadbe ),

    req( 'rd', 0xa, base_addr+0, 0, 0          ), resp( 'rd', 0xa, 0, 0xdeadbeef ),

  ]

#----------------------------------------------------------------------
# Test Case: subword writes
#----------------------------------------------------------------------

def subword_wr_msgs( base_addr ):
  return [

    req( 'wr', 0x0, base_addr+0, 1, 0x000000ef ), resp( 'wr', 0x0, 0, 0          ),
    req( 'wr', 0x1, base_addr+1, 1, 0x000000be ), resp( 'wr', 0x1, 0, 0          ),
    req( 'wr', 0x2, base_addr+2, 1, 0x000000ad ), resp( 'wr', 0x2, 0, 0          ),
    req( 'wr', 0x3, base_addr+3, 1, 0x000000de ), resp( 'wr', 0x3, 0, 0          ),
    req( 'rd', 0x4, base_addr+0, 0, 0          ), resp( 'rd', 0x4, 0, 0xdeadbeef ),

    req( 'wr', 0x5, base_addr+0, 2, 0x0000abcd ), resp( 'wr', 0x5, 0, 0          ),
    req( 'wr', 0x6, base_addr+2, 2, 0x0000ef01 ), resp( 'wr', 0x6, 0, 0          ),
    req( 'rd', 0x7, base_addr+0, 0, 0          ), resp( 'rd', 0x7, 0, 0xef01abcd ),

    req( 'wr', 0x8, base_addr+1, 2, 0x00002345 ), resp( 'wr', 0x8, 0, 0          ),
    req( 'rd', 0xa, base_addr+0, 0, 0          ), resp( 'rd', 0xa, 0, 0xef2345cd ),

    req( 'wr', 0xb, base_addr+0, 3, 0x00cafe02 ), resp( 'wr', 0xb, 0, 0          ),
    req( 'rd', 0xc, base_addr+0, 0, 0          ), resp( 'rd', 0xc, 0, 0xefcafe02 ),

  ]

#----------------------------------------------------------------------
# Test Case: amos
#----------------------------------------------------------------------

def amo_msgs( base_addr ):
  return [
    # load some initial data
    req( 'wr', 0x0, base_addr   , 0, 0x01234567 ), resp( 'wr', 0x0, 0, 0          ),
    req( 'wr', 0x0, base_addr+4 , 0, 0x98765432 ), resp( 'wr', 0x0, 0, 0          ),
    req( 'wr', 0x0, base_addr+8 , 0, 0x22002200 ), resp( 'wr', 0x0, 0, 0          ),
    req( 'wr', 0x0, base_addr+12, 0, 0x00112233 ), resp( 'wr', 0x0, 0, 0          ),
    req( 'wr', 0x0, base_addr+16, 0, 0x44556677 ), resp( 'wr', 0x0, 0, 0          ),
    req( 'wr', 0x0, base_addr+20, 0, 0x01230123 ), resp( 'wr', 0x0, 0, 0          ),
    # amo.add
    req( 'ad', 0x1, base_addr   , 0, 0x12345678 ), resp( 'ad', 0x1, 0, 0x01234567 ),
    req( 'rd', 0x2, base_addr   , 0, 0          ), resp( 'rd', 0x2, 0, 0x13579bdf ),
    # amo.and
    req( 'an', 0x3, base_addr+4 , 0, 0x23456789 ), resp( 'an', 0x3, 0, 0x98765432 ),
    req( 'rd', 0x4, base_addr+4 , 0, 0          ), resp( 'rd', 0x4, 0, 0x00444400 ),
    # amo.or
    req( 'or', 0x5, base_addr+8 , 0, 0x01230123 ), resp( 'or', 0x5, 0, 0x22002200 ),
    req( 'rd', 0x6, base_addr+8 , 0, 0          ), resp( 'rd', 0x6, 0, 0x23232323 ),
    # amo.xchg
    req( 'xg', 0x5, base_addr+12, 0, 0xdeadbeef ), resp( 'xg', 0x5, 0, 0x00112233 ),
    req( 'rd', 0x6, base_addr+12, 0, 0          ), resp( 'rd', 0x6, 0, 0xdeadbeef ),
    # amo.min -- mem is smaller
    req( 'mn', 0x7, base_addr+16, 0, 0xcafebabe ), resp( 'mn', 0x7, 0, 0x44556677 ),
    req( 'rd', 0x8, base_addr+16, 0, 0          ), resp( 'rd', 0x8, 0, 0xcafebabe ),
    # amo.min -- arg is smaller
    req( 'mn', 0x9, base_addr+20, 0, 0x01201234 ), resp( 'mn', 0x9, 0, 0x01230123 ),
    req( 'rd', 0xa, base_addr+20, 0, 0          ), resp( 'rd', 0xa, 0, 0x01201234 ),
  ]

#----------------------------------------------------------------------
# Test Case: random
#----------------------------------------------------------------------

def random_msgs( base_addr ):

  rgen = random.Random()
  rgen.seed(0xa4e28cc2)

  vmem = [ rgen.randint(0,0xffffffff) for _ in range(20) ]
  msgs = []

  for i in range(20):
    msgs.extend([
      req( 'wr', i, base_addr+4*i, 0, vmem[i] ), resp( 'wr', i, 0, 0 ),
    ])

  for i in range(1):
    idx = rgen.randint(0,19)

    if rgen.randint(0,1):

      correct_data = vmem[idx]
      msgs.extend([
        req( 'rd', i, base_addr+4*idx, 0, 0 ), resp( 'rd', i, 0, correct_data ),
      ])

    else:

      new_data = rgen.randint(0,0xffffffff)
      vmem[idx] = new_data
      msgs.extend([
        req( 'wr', i, base_addr+4*idx, 0, new_data ), resp( 'wr', i, 0, 0 ),
      ])

  return msgs

#-------------------------------------------------------------------------
# Test Case Table
#-------------------------------------------------------------------------

test_case_table = mk_test_case_table([
  (                             "msg_func          stall lat src_init src_intv sink_init sink_intv"),
  [ "basic",                     basic_msgs,       0.0,  1,  0,       0,       0,        0         ],
  [ "stream",                    stream_msgs,      0.0,  1,  0,       0,       0,        0         ],
  [ "subword_rd",                subword_rd_msgs,  0.0,  1,  0,       0,       0,        0         ],
  [ "subword_wr",                subword_wr_msgs,  0.0,  1,  0,       0,       0,        0         ],
  [ "amo",                       amo_msgs,         0.0,  1,  0,       0,       0,        0         ],
  [ "random",                    random_msgs,      0.0,  1,  0,       0,       0,        0         ],
  [ "random_3x14",               random_msgs,      0.0,  1,  5,       3,       7,        14        ],
  [ "stream_stall0.5_lat0",      stream_msgs,      0.5,  1,  0,       0,       0,        0         ],
  [ "stream_stall0.0_lat4",      stream_msgs,      0.0,  4,  0,       0,       0,        0         ],
  [ "stream_stall0.5_lat4",      stream_msgs,      0.5,  4,  0,       0,       0,        0         ],
  [ "random_stall0.5_lat4_3x14", random_msgs,      0.5,  4,  5,       14,      7,        14        ],
])

#-------------------------------------------------------------------------
# Test cases for 1 port
#-------------------------------------------------------------------------

#  @pytest.mark.parametrize( **test_case_table )
#  def test_1port( test_params, dump_vcd ):
  #  msgs = test_params.msg_func(0x1000)
  #  run_sim( TestHarness( 1, [ msgs[::2] ], [ msgs[1::2] ],
                        #  test_params.stall, test_params.lat,
                        #  test_params.src, test_params.sink ),
           #  dump_vcd )

#-------------------------------------------------------------------------
# Test cases for 2 port
#-------------------------------------------------------------------------

@pytest.mark.parametrize( **test_case_table )
def test_2port( test_params, dump_vcd ):
  msgs0 = test_params.msg_func(0x1000)
  msgs1 = test_params.msg_func(0x2000)
  run_sim( TestHarness( MemoryCL, 2, [(req_cls, resp_cls)]*2,
                        [ msgs0[::2],  msgs1[::2]  ],
                        [ msgs0[1::2], msgs1[1::2] ],
                        test_params.stall, test_params.lat,
                        test_params.src_init, test_params.src_intv,
                        test_params.sink_init, test_params.sink_intv ) )

@pytest.mark.parametrize( **test_case_table )
def test_20port( test_params, dump_vcd ):
  msgs = [ test_params.msg_func(0x1000*i) for i in range(20) ]
  run_sim( TestHarness( MemoryCL, 20, [(req_cls, resp_cls)]*20,
                        [ x[::2]  for x in msgs ],
                        [ x[1::2] for x in msgs ],
                        test_params.stall, test_params.lat,
                        test_params.src_init, test_params.src_intv,
                        test_params.sink_init, test_params.sink_intv ) )
#-------------------------------------------------------------------------
# Test Read/Write Mem
#-------------------------------------------------------------------------

def test_read_write_mem( dump_vcd ):

  rgen = random.Random()
  rgen.seed(0x05a3e95b)

  # Test data we want to write into memory

  data = [ rgen.randrange(-(2**31),2**31) for _ in range(20) ]

  # Convert test data into byte array

  data_bytes = struct.pack("<{}i".format(len(data)),*data)

  # Create memory messages to read and verify memory contents

  msgs = []
  for i, item in enumerate(data):
    msgs.extend([
      req( 'rd', 0x1, 0x1000+4*i, 0, 0 ), resp( 'rd', 0x1, 0, item ),
    ])

  # Create test harness with above memory messages

  th = TestHarness( MemoryCL, 2, [(req_cls, resp_cls)]*2, [msgs[::2], []], [msgs[1::2], []],
                    0, 0, 0, 0, 0, 0 )
  th.elaborate()

  # Write the data into the test memory

  th.mem.write_mem( 0x1000, data_bytes )

  # Run the test

  run_sim( th )

  # Read the data back out of the test memory

  result_bytes = th.mem.read_mem( 0x1000, len(data_bytes) )

  # Convert result bytes into list of ints

  result = list(struct.unpack("<{}i".format(len(data)),result_bytes))

  # Compare result to original data

  assert result == data

def run_sim( th, max_cycles=1000 ):

  # Create a simulator

  th.apply( SimulationPass() )
  th.sim_reset()

  # Run simulation

  print("")
  ncycles = 0
  print("{:3}:{}".format( ncycles, th.line_trace() ))
  while not th.done() and ncycles < max_cycles:
    th.tick()
    ncycles += 1
    print("{:3}:{}".format( ncycles, th.line_trace() ))

  # Check timeout

  assert ncycles < max_cycles

  th.tick()
  th.tick()
  th.tick()



"""
========================================================================
PrintWavePass.py
========================================================================
Print single bit signal in wave form and multi-bit signal by showing
least significant bits.

To use, call top.print_wave()

Inspired by PyRTL's state machine screenshot, which shows the change of signal
values along ticks of the clock.

Link: https://ucsbarchlab.github.io/PyRTL/

Author : Kaishuo Cheng, Shunning Jiang
Date   : Nov 9, 2019
"""

import py

from pymtl3.dsl import Const
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError


class PrintWavePass( BasePass ):

  def __call__( self, top ):
    if hasattr( top, "config_tracing" ):
      top.config_tracing.check()

      if top.config_tracing.tracing in [ 'text_ascii', 'text_fancy' ]:
        if not hasattr( top._tracing, "text_sigs" ):
          raise PassOrderError( "text_sigs" )

        def gen_print_wave( top ):
          def print_wave():
            _help_print( top )
          return print_wave

        top.print_textwave = gen_print_wave(top)

def _process_binary( sig, base, max ):
  """
  Returns int value from a signal in 32b form. Used for testing.

  Example: input: 0b00000000000000000000000000000000 10
           output: 0
  """
  if sig[1] == "b":
    sig = sig[2:]
  if base == 10:
    temp_int = int(sig,2)
    if sig[0] == '1':
      return temp_int -2 **32     # taking 2's complement.
                                  # leading 1 indicates a negative number
    else:
      return temp_int
    #hex number
  else:
    temp_hex = hex(int(sig,2))[2:]
    l = len(temp_hex)
    if l > max:
      temp_hex = temp_hex[l-max:]

    if l < max:
      temp_hex = '0'*(max-l) + temp_hex
    return temp_hex

def _help_print( self ):
  #default is text_fancy
  char_length = 5
  tick = '\u258f'
  up, down = '\u2571', '\u2572'
  x, low, high = '\u2573', '\u005f', '\u203e'
  revstart, revstop = '\x1B[7m', '\x1B[0m'
  light_gray = '\033[47m'
  back='\033[0m'  #back to normal printing
  if self.config_tracing.tracing == 'text_ascii':
    up,down,x,low = '/','\\','X','_'
    tick ='|'
  all_signals = self._tracing.text_sigs
  #spaces before cycle number
  max_length = 0
  for sig in all_signals:
    #   Example: s.in(12b)
    # length of signal name + (b) + number of digits, like 12
    # to add(32b) in front, add this to maxlength:
    #len(str(len(all_signals[sig][0][0])))+3
    this_length = len(sig)
    if this_length > max_length:
       if sig != "s.clk" and sig != "s.reset":
          max_length = this_length

  print()
  print()
  print(" "*(max_length+1),end = "")

  #handles clock tick symbol
  for i in range(0,len(all_signals["s.clk"])):
    #insert a space every 5 cycles
    if i % 5 == 0:
      if i != 0:
        print(" ",end ="")
      print(tick + str(i)+ " "*(char_length-1-len(str(i))) ,end="")
    else:
      print(tick + " "*(char_length-1) ,end="")


  print("")
  #signals
  for sig in sorted(all_signals.keys()):
    if sig != "s.clk" and sig != "s.reset":
      print("")
      bit_length = len(all_signals[sig][0])-2
      pos = sig.find('.')
      suffix = ""   # once used for (32b)
      #print suffix
      print((sig[pos+1:]+suffix).rjust(max_length),end="")
      prev_sig = None
      # one bit
      if bit_length==1:
        for i, val in enumerate( all_signals[sig] ):

          #every 5 cycles add a space
          if i%5 == 0:
            print(" ",end = "")
          if val[2] == '1':
            current_sig = high
          else:
            current_sig = low
          # detecting if first cycle
          if prev_sig is not None:
            if prev_sig == low and current_sig == high:
              print(up+current_sig*(char_length-1),end = "")
            elif prev_sig == high and current_sig == low:
              print(down+current_sig*(char_length-1),end = "")
            # prev and current signal agree
            else:
              print(current_sig*char_length,end = "")
          # first cycle
          else:
            print(current_sig*char_length,end = "")
          prev_sig = current_sig

        print("")
        #multiple bits
      else:
        next = 0
        val_list = all_signals[sig]
        for i in range(len(val_list)):
          # signals in this cycle is still the same as before
          if next > 0:
            next -=1
            continue

          val = val_list[i]
          for j in range(i,len(val_list)):
            # a space every 5 cycles

            if val_list[j] != val:
              j = j-1
              break
            if j % 5 == 4:
              break
          if i%5 == 0:
            print(back +" ",end = "")

          #first is reserved for X or " ". Rest is 5 char length.
          length = 4+ 5*(j-i)
          next = j-i
          # print(sig,end = " ")
          # print(j, end = " ")
          # print(i)
          if length >= bit_length//4:
            length = bit_length//4
            if bit_length % 4 != 0:
              length+=1
            plus = False
          else:
            #reverse a place for +
            length = length -1
            plus = True

          current = _process_binary(val,16,length)
          # print a +, with one less space for signal number
          if plus:
            if i==0:
              print(light_gray + " " +'\033[30m'+"+"+ current,end = "")
            else:
              print(light_gray + '\033[30m'+x + "+" +current,end = "")
          # no +, more spaces for signal number
          else:
            if i==0:
              print(light_gray + " " +'\033[30m'+ current,end = "")
            else:
              print(light_gray + '\033[30m'+x +current,end = "")
            print(" "*(4+5*(j-i)-length),end = "")

        print(back + "")

  print("")



#=========================================================================
# TracingConfigs.py
#=========================================================================
# Author : Shunning Jiang
# Date   : Dec 4, 2019
"""Configuration class of tracing passes"""


from pymtl3.passes.PassConfigs import BasePassConfigs, Checker


class TracingConfigs( BasePassConfigs ):

  Options = {
    "tracing" : 'none',
    "vcd_file_name" : "",
    "method_trace" : True,
  }

  Checkers = {
    'tracing': Checker(
      lambda v: v in ['none', 'vcd', 'text_ascii', 'text_fancy' ],
      "expects a string in ['none', 'vcd', 'text_ascii', 'text_fancy']"
    ),

    'vcd_file_name': Checker(
      condition = lambda v: isinstance(v, str),
      error_msg = "expects a string"
    ),

    'method_trace': Checker(
      condition = lambda v: isinstance(v, bool),
      error_msg = "expects a boolean"
    ),
  }

  PassName = 'passes.tracing.*'



"""
========================================================================
Test sources
========================================================================
Test sources with CL or RTL interfaces.

Author : Yanghui Ou
  Date : Mar 11, 2019
"""
from collections import deque

from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvCL2SendRTL, SendIfcRTL

#-------------------------------------------------------------------------
# TestSrcCL
#-------------------------------------------------------------------------

class TestSrcCL( Component ):

  def construct( s, Type, msgs, initial_delay=0, interval_delay=0 ):

    s.send = CallerIfcCL( Type=Type )
    s.msgs = deque( msgs )

    s.count  = initial_delay
    s.delay  = interval_delay

    @s.update
    def up_src_send():
      if s.count > 0:
        s.count -= 1
      elif not s.reset:
        if s.send.rdy() and s.msgs:
          s.send( s.msgs.popleft() )
          s.count = s.delay # reset count after a message is sent

  def done( s ):
    return not s.msgs

  # Line trace

  def line_trace( s ):
    return "{}".format( s.send )

#-------------------------------------------------------------------------
# TestSrcRTL
#-------------------------------------------------------------------------
# TODO: deprecating TestSrcRTL.

class TestSrcRTL( Component ):

  def construct( s, Type, msgs, initial_delay=0, interval_delay=0 ):

    # Interface

    s.send = SendIfcRTL( Type )

    # Components

    s.src     = TestSrcCL( Type, msgs, initial_delay, interval_delay )
    s.adapter = RecvCL2SendRTL( Type )

    connect( s.src.send,     s.adapter.recv )
    connect( s.adapter.send, s.send         )

  def done( s ):
    return s.src.done()

  # Line trace

  def line_trace( s ):
    return "{}".format( s.send )



#=========================================================================
# BehavioralRTLIRTypeCheckL1Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 20, 2019
"""Provide L1 behavioral RTLIR type check pass."""

import copy
from collections import OrderedDict

from pymtl3.datatypes import Bits32, mk_bits
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError, RTLIRConversionError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from . import BehavioralRTLIR as bir


class BehavioralRTLIRTypeCheckL1Pass( BasePass ):
  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    if not hasattr( m, '_pass_behavioral_rtlir_type_check' ):
      m._pass_behavioral_rtlir_type_check = PassMetadata()
    m._pass_behavioral_rtlir_type_check.rtlir_freevars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_accessed = set()
    visitor = BehavioralRTLIRTypeCheckVisitorL1(
      m, m._pass_behavioral_rtlir_type_check.rtlir_freevars,
      m._pass_behavioral_rtlir_type_check.rtlir_accessed )
    for blk in m.get_update_block_order():
      visitor.enter( blk, m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )

class BehavioralRTLIRTypeCheckVisitorL1( bir.BehavioralRTLIRNodeVisitor ):
  def __init__( s, component, freevars, accessed ):
    s.component = component
    s.freevars = freevars
    s.accessed = accessed
    s.type_expect = {}
    lhs_types = ( rt.Port, rt.Wire, rt.NetWire )
    index_types = ( rt.Port, rt.Wire, rt.Array )
    slice_types = ( rt.Port, rt.Wire )
    s.type_expect[ 'Assign' ] = {
      'target' : ( lhs_types, 'lhs of assignment must be a signal!' ),
      'value' : ( rt.Signal, 'rhs of assignment should be signal or const!' )
    }
    s.type_expect[ 'ZeroExt' ] = {
      'value':( rt.Signal, 'extension only applies to signals!' )
    }
    s.type_expect[ 'SignExt' ] = {
      'value':( rt.Signal, 'extension only applies to signals!' )
    }
    s.type_expect[ 'Reduce' ] = {
      'value':( rt.Signal, 'reduce only applies on a signal!' )
    }
    s.type_expect[ 'SizeCast' ] = {
      'value':( rt.Signal, 'size casting only applies to signals/consts!' )
    }
    s.type_expect[ 'Attribute' ] = {
      'value':( rt.Component, 'the base of an attribute must be a component!' )
    }
    s.type_expect[ 'Index' ] = {
      'idx':( rt.Signal, 'index must be a signal or constant expression!' ),
      'value':( index_types, 'the base of an index must be an array or signal!' )
    }
    s.type_expect[ 'Slice' ] = {
      'value':( slice_types, 'the base of a slice must be a signal!' ),
      'lower':( rt.Signal, 'upper of slice must be a constant expression!' ),
      'upper':( rt.Signal, 'lower of slice must be a constant expression!' )
    }

  def enter( s, blk, rtlir ):
    """ entry point for RTLIR type checking """
    s.blk     = blk

    # s.globals contains a dict of the global namespace of the module where
    # blk was defined
    s.globals = blk.__globals__

    # s.closure contains the free variables defined in an enclosing scope.
    # Basically this is the model instance s.
    s.closure = {}

    for i, var in enumerate( blk.__code__.co_freevars ):
      try:
        s.closure[ var ] = blk.__closure__[ i ].cell_contents
      except ValueError:
        pass
    s.visit( rtlir )

  # Override the default visit()
  def visit( s, node ):
    node_name = node.__class__.__name__
    method = 'visit_' + node_name
    func = getattr( s, method, s.generic_visit )

    # First visit (type check) all child nodes
    for field, value in vars(node).items():
      if isinstance( value, list ):
        for item in value:
          if isinstance( item, bir.BaseBehavioralRTLIR ):
            s.visit( item )
      elif isinstance( value, bir.BaseBehavioralRTLIR ):
        s.visit( value )

    # Then verify that all child nodes have desired types
    try:
      # Check the expected types of child nodes
      for field, type_rule in s.type_expect[node_name].items():
        value = vars(node)[field]
        target_type = type_rule[ 0 ]
        exception_msg = type_rule[ 1 ]
        if eval( 'not isinstance( value.Type, target_type )' ):
          raise PyMTLTypeError( s.blk, node.ast, exception_msg )
    except PyMTLTypeError:
      raise
    except Exception:
      # This node does not require type checking on child nodes
      pass

    # Finally call the type check function
    func( node )

  # Override the default generic_visit()
  def generic_visit( s, node ):
    node.Type = None

  def is_same( s, u, v ):
    """Return if the sub-AST at u and v are the same."""
    if type(u) is not type(v):
      return False
    return u == v

  def visit_Assign( s, node ):
    # RHS should have the same type as LHS
    rhs_type = node.value.Type.get_dtype()
    lhs_type = node.target.Type.get_dtype()

    # Weak type checking (agreeable types)
    if not lhs_type( rhs_type ):
      raise PyMTLTypeError( s.blk, node.ast,
        f'Unagreeable types {lhs_type} and {rhs_type}!' )

    # Strong type checking (same type)
    if rhs_type != lhs_type:
      raise PyMTLTypeError( s.blk, node.ast,
        f'LHS and RHS of assignment should have the same type ({lhs_type} vs {rhs_type})!' )

    node.Type = None

  def visit_FreeVar( s, node ):
    if node.name not in s.freevars:
      s.freevars[ node.name ] = node.obj

    try:
      t = rt.get_rtlir( node.obj )
    except RTLIRConversionError as e:
      raise PyMTLTypeError(s.blk, node.ast,
        f'{node.name} cannot be converted into a valid RTLIR object!' )

    if isinstance( t, rt.Const ) and isinstance( t.get_dtype(), rdt.Vector ):
      node._value = mk_bits( t.get_dtype().get_length() )( node.obj )
    node.Type = t

  def visit_Base( s, node ):
    # Mark this node as having type rt.Component
    # In L1 the `s` top component is the only possible base
    node.Type = rt.get_rtlir( node.base )
    if not isinstance( node.Type, rt.Component ):
      raise PyMTLTypeError( s.blk, node.ast,
        f'{node} is not a rt.Component!' )

  def visit_Number( s, node ):
    # By default, number literals have bitwidth of 32
    node.Type = rt.get_rtlir( node.value )
    node._value = Bits32( node.value )

  def visit_Concat( s, node ):
    nbits = 0
    for child in node.values:
      if not isinstance(child.Type, rt.Signal):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{child} is not a signal!' )
      nbits += child.Type.get_dtype().get_length()
    node.Type = rt.NetWire( rdt.Vector( nbits ) )

  def visit_ZeroExt( s, node ):
    try:
      new_nbits = node.nbits._value
    except AttributeError:
      raise PyMTLTypeError( s.blk, node.ast,
        f'{node.nbits} is not a constant number!' )
    child_type = node.value.Type
    old_nbits = child_type.get_dtype().get_length()
    if new_nbits <= old_nbits:
      raise PyMTLTypeError( s.blk, node.ast,
        f'{new_nbits} is not greater than {old_nbits}!' )
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( new_nbits )

  def visit_SignExt( s, node ):
    try:
      new_nbits = node.nbits._value
    except AttributeError:
      raise PyMTLTypeError( s.blk, node.ast,
        f'{node.nbits} is not a constant number!' )
    child_type = node.value.Type
    old_nbits = child_type.get_dtype().get_length()
    if new_nbits <= old_nbits:
      raise PyMTLTypeError( s.blk, node.ast,
        f'{new_nbits} is not greater than {old_nbits}!' )
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( new_nbits )

  def visit_Reduce( s, node ):
    child_type = node.value.Type
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( 1 )

  def visit_SizeCast( s, node ):
    nbits = node.nbits
    Type = node.value.Type

    # We do not check for bitwidth mismatch here because the user should
    # be able to explicitly convert signals/constatns to different bitwidth.
    node.Type = copy.copy( Type )
    node.Type.dtype = rdt.Vector( nbits )

    try:
      node._value = node.value._value
    except AttributeError:
      pass

  def visit_Attribute( s, node ):
    # Attribute supported at L1: CurCompAttr
    if isinstance( node.value, bir.Base ):
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'type {node.value.Type} does not have attribute {node.attr}!' )
      s.accessed.add( node.attr )

    else:
      raise PyMTLTypeError( s.blk, node.ast,
        f'non-component attribute {node.attr} of type {node.value.Type} is not supported at L1!' )
    # value.attr has the type that is specified by the base
    node.Type = node.value.Type.get_property( node.attr )

  def visit_Index( s, node ):
    idx = None if not hasattr(node.idx, "_value") else node.idx._value
    if isinstance( node.value.Type, rt.Array ):
      if idx is not None and not (0 <= idx < node.value.Type.get_dim_sizes()[0]):
        raise PyMTLTypeError( s.blk, node.ast, 'array index out of range!' )
      node.Type = node.value.Type.get_next_dim_type()
      obj = node.value.Type.get_obj()
      if idx is not None and obj is not None:
        if isinstance( node.Type, rt.Array ):
          node.Type.obj = obj[ int( idx ) ]
        else:
          node._value = obj[ int( idx ) ]

    elif isinstance( node.value.Type, rt.Signal ):
      dtype = node.value.Type.get_dtype()
      if node.value.Type.is_packed_indexable():
        if idx is not None and not (0 <= idx < dtype.get_length()):
          raise PyMTLTypeError( s.blk, node.ast,
            'bit selection index out of range!' )
        node.Type = node.value.Type.get_next_dim_type()
      elif isinstance( dtype, rdt.Vector ):
        if idx is not None and not(0 <= idx < dtype.get_length()):
          raise PyMTLTypeError( s.blk, node.ast,
            'bit selection index out of range!' )
        node.Type = rt.NetWire( rdt.Vector( 1 ) )
      else:
        raise PyMTLTypeError( s.blk, node.ast,
          f'cannot perform index on {dtype}!')

    else:
      # Should be unreachable
      raise PyMTLTypeError( s.blk, node.ast,
        f'cannot perform index on {node.value.Type}!')

  def visit_Slice( s, node ):
    lower_val = None if not hasattr(node.lower, "_value") else node.lower._value
    upper_val = None if not hasattr(node.upper, "_value") else node.upper._value
    dtype = node.value.Type.get_dtype()

    if not isinstance( dtype, rdt.Vector ):
      raise PyMTLTypeError( s.blk, node.ast, f'cannot perform slicing on type {dtype}!')

    if not lower_val is None and not upper_val is None:
      signal_nbits = dtype.get_length()
      # upper bound must be strictly larger than the lower bound
      if ( lower_val >= upper_val ):
        raise PyMTLTypeError( s.blk, node.ast,
          'the upper bound of a slice must be larger than the lower bound!' )
      # upper & lower bound should be less than the bit width of the signal
      if not ( 0 <= lower_val < upper_val <= signal_nbits ):
        raise PyMTLTypeError( s.blk, node.ast,
          'upper/lower bound of slice out of width of signal!' )
      node.Type = rt.NetWire( rdt.Vector( int( upper_val - lower_val ) ) )

    else:
      # Try to special case the constant-stride part selection
      try:
        assert isinstance( node.upper, bir.BinOp )
        assert isinstance( node.upper.op, bir.Add )
        nbits = node.upper.right
        slice_size = nbits._value
        assert s.is_same( node.lower, node.upper.left )
        node.Type = rt.NetWire( rdt.Vector( slice_size ) )
        # Add new fields that might help translation
        node.size = slice_size
        node.base = node.lower
      except Exception:
        raise PyMTLTypeError( s.blk, node.ast, 'slice bounds must be constant!' )



#=========================================================================
# BehavioralRTLIR.py
#=========================================================================
"""Provide behavioral RTLIR AST node types.

This file is automatically generated by BehavioralRTLIRImplGen.py.
"""

class BaseBehavioralRTLIR:
  """Base class for all behavioral RTLIR AST nodes."""
  def __eq__( s, other ):
    return type(s) is type(other)

  def __ne__( s, other ):
    return not s.__eq__( other )

class CombUpblk( BaseBehavioralRTLIR ):
  def __init__( s, name, body ):
    s.name = name
    s.body = body

  def __eq__( s, other ):
    if not isinstance(other, CombUpblk) or s.name != other.name:
      return False
    for x, y in zip( s.body, other.body ):
      if x != y:
        return False
    return True

class SeqUpblk( BaseBehavioralRTLIR ):
  def __init__( s, name, body ):
    s.name = name
    s.body = body

  def __eq__( s, other ):
    if not isinstance(other, SeqUpblk) or s.name != other.name:
      return False
    for x, y in zip( s.body, other.body ):
      if x != y:
        return False
    return True

class Assign( BaseBehavioralRTLIR ):
  def __init__( s, target, value, blocking ):
    s.target = target
    s.value = value
    s.blocking = blocking

  def __eq__( s, other ):
    return isinstance(other, Assign) and s.target == other.target and s.value == other.value and s.blocking == other.blocking

class If( BaseBehavioralRTLIR ):
  def __init__( s, cond, body, orelse ):
    s.cond = cond
    s.body = body
    s.orelse = orelse

  def __eq__( s, other ):
    if not isinstance(other, If) or s.cond != other.cond:
      return False
    for x, y in zip( s.body, other.body ):
      if x != y:
        return False
    for x, y in zip( s.orelse, other.orelse ):
      if x != y:
        return False
    return True

class For( BaseBehavioralRTLIR ):
  def __init__( s, var, start, end, step, body ):
    s.var = var
    s.start = start
    s.end = end
    s.step = step
    s.body = body

  def __eq__( s, other ):
    if not isinstance(other, For) or s.var != other.var or s.start != other.start or s.end != other.end or s.step != other.step:
      return False
    for x, y in zip( s.body, other.body ):
      if x != y:
        return False
    return True

class Number( BaseBehavioralRTLIR ):
  def __init__( s, value ):
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, Number) and s.value == other.value

class Concat( BaseBehavioralRTLIR ):
  def __init__( s, values ):
    s.values = values

  def __eq__( s, other ):
    if not isinstance(other, Concat):
      return False
    for x, y in zip( s.values, other.values ):
      if x != y:
        return False
    return True

class ZeroExt( BaseBehavioralRTLIR ):
  def __init__( s, nbits, value ):
    s.nbits = nbits
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, ZeroExt) and s.nbits == other.nbits and s.value == other.value

class SignExt( BaseBehavioralRTLIR ):
  def __init__( s, nbits, value ):
    s.nbits = nbits
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, SignExt) and s.nbits == other.nbits and s.value == other.value

class Reduce( BaseBehavioralRTLIR ):
  def __init__( s, op, value ):
    s.op = op
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, Reduce) and s.op == other.op and s.value == other.value

class SizeCast( BaseBehavioralRTLIR ):
  def __init__( s, nbits, value ):
    s.nbits = nbits
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, SizeCast) and s.nbits == other.nbits and s.value == other.value

class StructInst( BaseBehavioralRTLIR ):
  def __init__( s, struct, values ):
    s.struct = struct
    s.values = values

  def __eq__( s, other ):
    if not isinstance(other, StructInst) or s.struct != other.struct:
      return False
    for x, y in zip( s.values, other.values ):
      if x != y:
        return False
    return True

class IfExp( BaseBehavioralRTLIR ):
  def __init__( s, cond, body, orelse ):
    s.cond = cond
    s.body = body
    s.orelse = orelse

  def __eq__( s, other ):
    return isinstance(other, IfExp) and s.cond == other.cond and s.body == other.body and s.orelse == other.orelse

class UnaryOp( BaseBehavioralRTLIR ):
  def __init__( s, op, operand ):
    s.op = op
    s.operand = operand

  def __eq__( s, other ):
    return isinstance(other, UnaryOp) and s.op == other.op and s.operand == other.operand

class BoolOp( BaseBehavioralRTLIR ):
  def __init__( s, op, values ):
    s.op = op
    s.values = values

  def __eq__( s, other ):
    if not isinstance(other, BoolOp) or s.op != other.op:
      return False
    for x, y in zip( s.values, other.values ):
      if x != y:
        return False
    return True

class BinOp( BaseBehavioralRTLIR ):
  def __init__( s, left, op, right ):
    s.left = left
    s.op = op
    s.right = right

  def __eq__( s, other ):
    return isinstance(other, BinOp) and s.left == other.left and s.op == other.op and s.right == other.right

class Compare( BaseBehavioralRTLIR ):
  def __init__( s, left, op, right ):
    s.left = left
    s.op = op
    s.right = right

  def __eq__( s, other ):
    return isinstance(other, Compare) and s.left == other.left and s.op == other.op and s.right == other.right

class Attribute( BaseBehavioralRTLIR ):
  def __init__( s, value, attr ):
    s.value = value
    s.attr = attr

  def __eq__( s, other ):
    return isinstance(other, Attribute) and s.value == other.value and s.attr == other.attr

class Index( BaseBehavioralRTLIR ):
  def __init__( s, value, idx ):
    s.value = value
    s.idx = idx

  def __eq__( s, other ):
    return isinstance(other, Index) and s.value == other.value and s.idx == other.idx

class Slice( BaseBehavioralRTLIR ):
  def __init__( s, value, lower, upper, base = None, size = None ):
    s.value = value
    s.lower = lower
    s.upper = upper
    s.base = base
    s.size = size

  def __eq__( s, other ):
    return isinstance(other, Slice) and s.value == other.value and s.lower == other.lower and s.upper == other.upper and s.base == other.base and s.size == other.size

class Base( BaseBehavioralRTLIR ):
  def __init__( s, base ):
    s.base = base

  def __eq__( s, other ):
    return isinstance(other, Base) and s.base == other.base

class LoopVar( BaseBehavioralRTLIR ):
  def __init__( s, name ):
    s.name = name

  def __eq__( s, other ):
    return isinstance(other, LoopVar) and s.name == other.name

class FreeVar( BaseBehavioralRTLIR ):
  def __init__( s, name, obj ):
    s.name = name
    s.obj = obj

  def __eq__( s, other ):
    return isinstance(other, FreeVar) and s.name == other.name and s.obj == other.obj

class TmpVar( BaseBehavioralRTLIR ):
  def __init__( s, name, upblk_name ):
    s.name = name
    s.upblk_name = upblk_name

  def __eq__( s, other ):
    return isinstance(other, TmpVar) and s.name == other.name and s.upblk_name == other.upblk_name

class LoopVarDecl( BaseBehavioralRTLIR ):
  def __init__( s, name ):
    s.name = name

  def __eq__( s, other ):
    return isinstance(other, LoopVarDecl) and s.name == other.name

class Invert( BaseBehavioralRTLIR ):
  pass

class Not( BaseBehavioralRTLIR ):
  pass

class UAdd( BaseBehavioralRTLIR ):
  pass

class USub( BaseBehavioralRTLIR ):
  pass

class And( BaseBehavioralRTLIR ):
  pass

class Or( BaseBehavioralRTLIR ):
  pass

class Add( BaseBehavioralRTLIR ):
  pass

class Sub( BaseBehavioralRTLIR ):
  pass

class Mult( BaseBehavioralRTLIR ):
  pass

class Div( BaseBehavioralRTLIR ):
  pass

class Mod( BaseBehavioralRTLIR ):
  pass

class Pow( BaseBehavioralRTLIR ):
  pass

class ShiftLeft( BaseBehavioralRTLIR ):
  pass

class ShiftRightLogic( BaseBehavioralRTLIR ):
  pass

class BitAnd( BaseBehavioralRTLIR ):
  pass

class BitOr( BaseBehavioralRTLIR ):
  pass

class BitXor( BaseBehavioralRTLIR ):
  pass

class Eq( BaseBehavioralRTLIR ):
  pass

class NotEq( BaseBehavioralRTLIR ):
  pass

class Lt( BaseBehavioralRTLIR ):
  pass

class LtE( BaseBehavioralRTLIR ):
  pass

class Gt( BaseBehavioralRTLIR ):
  pass

class GtE( BaseBehavioralRTLIR ):
  pass

class BehavioralRTLIRNodeVisitor:
  """Class for behavioral RTLIR AST visitors."""
  def visit( self, node, *args ):
    method = 'visit_' + node.__class__.__name__
    visitor = getattr( self, method, self.generic_visit )
    return visitor( node, *args )

  def generic_visit( self, node, *args ):
    for field, value in vars(node).items():
      if isinstance( value, list ):
        for item in value:
          if isinstance( item, BaseBehavioralRTLIR ):
            self.visit( item, *args )
      elif isinstance( value, BaseBehavioralRTLIR ):
        self.visit( value, *args )



"""
========================================================================
sim_utils.py
========================================================================
Simple util functions to bootstrap simulation without simulation passes.

Author : Shunning Jiang
Date   : Jan 1, 2018
"""
import random
from collections import defaultdict

import py.code

from pymtl3.datatypes import Bits1
from pymtl3.datatypes.helpers import get_bitstruct_inst_all_classes
from pymtl3.dsl.ComponentLevel1 import ComponentLevel1
from pymtl3.dsl.ComponentLevel2 import ComponentLevel2
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3
from pymtl3.dsl.ComponentLevel4 import ComponentLevel4
from pymtl3.dsl.ComponentLevel5 import ComponentLevel5
from pymtl3.dsl.ComponentLevel6 import ComponentLevel6
from pymtl3.dsl.ComponentLevel7 import ComponentLevel7
from pymtl3.dsl.Connectable import (
    BlockingIfc,
    Const,
    Interface,
    MethodPort,
    NonBlockingIfc,
    Signal,
)
from pymtl3.dsl.errors import (
    LeftoverPlaceholderError,
    NotElaboratedError,
    UpblkCyclicError,
)
from pymtl3.dsl.NamedObject import NamedObject
from pymtl3.dsl.Placeholder import Placeholder


def simple_sim_pass( s, seed=0xdeadbeef ):
  #  random.seed( seed )
  assert isinstance( s, ComponentLevel1 )

  if not hasattr( s._dsl, "all_U_U_constraints" ):
    raise NotElaboratedError()

  placeholders = [ x for x in s._dsl.all_named_objects
                   if isinstance( x, Placeholder ) ]

  if placeholders:
    raise LeftoverPlaceholderError( placeholders )

  all_upblks = set( s._dsl.all_upblks )
  expl_constraints = set( s._dsl.all_U_U_constraints )

  gen_upblk_reads  = {}
  gen_upblk_writes = {}

  if isinstance( s, ComponentLevel2 ):
    all_update_ff = set( s._dsl.all_update_ff )

    if isinstance( s, ComponentLevel3 ):
      nets = s.get_all_value_nets()

      for writer, signals in nets:
        if len(signals) == 1: continue
        readers = [ x for x in signals if x is not writer ]

        fanout  = len( readers )

        upblk_name = f"{writer!r}__{fanout}" \
                        .replace( ".", "_" ).replace( ":", "_" ) \
                        .replace( "[", "_" ).replace( "]", "" ) \
                        .replace( "(", "_" ).replace( ")", "" ) \
                        .replace( ",", "_" )

        rstrs    = [ f"{x!r} = _w" for x in readers ]
        _globals = { 's': s }

        if isinstance( writer, Const ) and type(writer._dsl.const) is not int:
          types = get_bitstruct_inst_all_classes( writer._dsl.const )

          for t in types:
            if t.__name__ in _globals:
              assert t is _globals[ t.__name__ ], "Cannot handle two subfields with the same struct name but different structs"
            _globals[ t.__name__ ] = t

        src = f"""
        def {upblk_name}():
          _w = {writer!r}
          {"; ".join(rstrs)}
        """
        _locals = {}
        exec(py.code.Source( src ).compile(), _globals, _locals)

        _recent_blk = _locals[upblk_name]

        # Collect read/writer metadata, directly insert them into _all_X

        all_upblks.add( _recent_blk )
        gen_upblk_reads [ _recent_blk ] = [ writer ]
        gen_upblk_writes[ _recent_blk ] = readers

    # s is ComponentLevel2

    #---------------------------------------------------------------------
    # Explicit constraint
    #---------------------------------------------------------------------
    # Schedule U1 before U2 when U1 == WR(x) < RD(x) == U2: combinational
    #
    # Explicitly, one should define these to invert the implicit constraint:
    # - RD(x) < U when U == WR(x) --> RD(x) ( == U') < U == WR(x)
    # - WR(x) > U when U == RD(x) --> RD(x) == U < WR(x) ( == U')
    # constraint RD(x) < U1 & U2 reads  x --> U2 == RD(x) <  U1
    # constraint RD(x) > U1 & U2 reads  x --> U1 <  RD(x) == U2 # impl
    # constraint WR(x) < U1 & U2 writes x --> U2 == WR(x) <  U1 # impl
    # constraint WR(x) > U1 & U2 writes x --> U1 <  WR(x) == U2
    # Doesn't work for nested data struct and slice:

    read_upblks = defaultdict(set)
    write_upblks = defaultdict(set)

    for data in [ s._dsl.all_upblk_reads, gen_upblk_reads ]:
      for blk, reads in data.items():
        for rd in reads:
          read_upblks[ rd ].add( blk )

    for data in [ s._dsl.all_upblk_writes, gen_upblk_writes ]:
      for blk, writes in data.items():
        for wr in writes:
          write_upblks[ wr ].add( blk )

    for typ in [ 'rd', 'wr' ]: # deduplicate code
      if typ == 'rd':
        constraints = s._dsl.all_RD_U_constraints
        equal_blks  = read_upblks
      else:
        constraints = s._dsl.all_WR_U_constraints
        equal_blks  = write_upblks

      # enumerate variable objects
      for obj, constrained_blks in constraints.items():

        # enumerate upblks that has a constraint with x
        for (sign, co_blk) in constrained_blks:

          for eq_blk in equal_blks[ obj ]: # blocks that are U == RD(x)
            if co_blk != eq_blk:
              if sign == 1: # RD/WR(x) < U is 1, RD/WR(x) > U is -1
                # eq_blk == RD/WR(x) < co_blk
                expl_constraints.add( (eq_blk, co_blk) )
              else:
                # co_blk < RD/WR(x) == eq_blk
                expl_constraints.add( (co_blk, eq_blk) )

    #---------------------------------------------------------------------
    # Implicit constraint
    #---------------------------------------------------------------------
    # Synthesize total constraints between two upblks that read/write to
    # the "same variable" (we also handle the read/write of a recursively
    # nested field/slice)
    #
    # Implicitly, WR(x) < RD(x), so when U1 writes X and U2 reads x
    # - U1 == WR(x) & U2 == RD(x) --> U1 == WR(x) < RD(x) == U2

    impl_constraints = set()

    # Collect all objs that write the variable whose id is "read"
    # 1) RD A.b.b     - WR A.b.b, A.b, A
    # 2) RD A.b[1:10] - WR A.b[1:10], A.b, A
    # 3) RD A.b[1:10] - WR A.b[0:5], A.b[6], A.b[8:11]

    for obj, rd_blks in read_upblks.items():
      writers = []

      # Check parents. Cover 1) and 2)
      x = obj
      while x.is_signal():
        if x in write_upblks:
          writers.append( x )
        x = x.get_parent_object()

      # Check the sibling slices. Cover 3)
      if obj.is_signal():
        for x in obj.get_sibling_slices():
          if x.slice_overlap( obj ) and x in write_upblks:
            writers.append( x )

      # Add all constraints
      for writer in writers:
        for wr_blk in write_upblks[ writer ]:
          if wr_blk not in all_update_ff:
            for rd_blk in rd_blks:
              if wr_blk != rd_blk:
                if rd_blk not in all_update_ff:
                  impl_constraints.add( (wr_blk, rd_blk) ) # wr < rd default

    # Collect all objs that read the variable whose id is "write"
    # 1) WR A.b.b.b, A.b.b, A.b, A (detect 2-writer conflict)
    # 2) WR A.b.b.b   - RD A.b.b, A.b, A
    # 3) WR A.b[1:10] - RD A.b[1:10], A,b, A
    # 4) WR A.b[1:10], A.b[0:5], A.b[6] (detect 2-writer conflict)
    # "WR A.b[1:10] - RD A.b[0:5], A.b[6], A.b[8:11]" has been discovered

    for obj, wr_blks in write_upblks.items():
      readers = []

      # Check parents. Cover 2) and 3). 1) and 4) should be detected in elaboration
      x = obj
      while x.is_signal():
        if x in read_upblks:
          readers.append( x )
        x = x.get_parent_object()

      # Add all constraints
      for wr_blk in wr_blks:
        if wr_blk not in all_update_ff:
          for reader in readers:
            for rd_blk in read_upblks[ reader ]:
              if wr_blk != rd_blk:
                if rd_blk not in all_update_ff:
                  impl_constraints.add( (wr_blk, rd_blk) ) # wr < rd default

    all_constraints = { *expl_constraints }
    for (x, y) in impl_constraints:
      if (y, x) not in expl_constraints: # no conflicting expl
        all_constraints.add( (x, y) )
  else:
    all_constraints = { *expl_constraints }

  #-----------------------------------------------------------------------
  # Process method constraints
  #----------------------------------------------------------------------
  # I assume method don't call other methods here

  # Do bfs to find out all potential total constraints associated with
  # each method, direction conflicts, and incomplete constraints

  verbose = False

  if isinstance( s, ComponentLevel4 ):
    method_blks = defaultdict(set)

    if isinstance( s, ComponentLevel5 ):
      for writer, net in s._dsl.all_method_nets:
        for member in net:
          if member is not writer:
            assert member.method is None
            member.method = writer.method

    # Collect each CalleePort/method is called in which update block
    # We use bounded method of CalleePort to identify each call
    for blk, calls in s._dsl.all_upblk_calls.items():
      if verbose: print("--", blk, calls)
      for call in calls:
        if isinstance( call, MethodPort ):
          method_blks[ call.method ].add( blk )
        elif isinstance( call, (NonBlockingIfc, BlockingIfc) ):
          method_blks[ call.method.method ].add( blk )
        else:
          method_blks[ call ].add( blk )

    # Put all M-related constraints into predecessor and successor dicts
    pred = defaultdict(set)
    succ = defaultdict(set)

    # We also pre-process M(x) == M(y) constraints into per-method
    # equivalence sets
    equiv = defaultdict(set)

    for (x, y, is_equal) in s._dsl.all_M_constraints:
      if verbose: print((x,y,is_equal))

      if   isinstance( x, MethodPort ):
        xx = x.method

      # We allow the user to call the interface directly in a non-blocking
      # interface, so if they do call it, we use the actual method within
      # the method field
      elif isinstance( x, (NonBlockingIfc, BlockingIfc) ):
        xx = x.method.method

      else:
        xx = x

      if   isinstance( y, MethodPort ):
        yy = y.method

      elif isinstance( y, (NonBlockingIfc, BlockingIfc) ):
        yy = y.method.method

      else:
        yy = y

      pred[ yy ].add( xx )
      succ[ xx ].add( yy )

      if is_equal: # M(x) == M(y)
        equiv[xx].add( yy )
        equiv[yy].add( xx )

    for method, assoc_blks in method_blks.items():
      visited = {  (method, 0)  }
      Q = [ (method, 0) ] # -1: pred, 0: don't know, 1: succ

      if verbose: print()
      while Q:
        (u, w) = Q.pop()
        if verbose: print((u, w))

        if u in equiv:
          for v in equiv[u]:
            if (v, w) not in visited:
              visited.add( (v, w) )
              Q.append( (v, w) )

        if w <= 0:
          for v in pred[u]:

            if v in all_upblks:
              # Find total constraint (v < blk) by v < method_u < method_u'=blk
              # INVALID if we have explicit constraint (blk < method_u)

              for blk in assoc_blks:
                if blk not in pred[u]:
                  if v != blk:
                    if verbose: print("w<=0, v is blk".center(10),v, blk)
                    if verbose: print(v.__name__.center(25)," < ", \
                                blk.__name__.center(25))
                    all_constraints.add( (v, blk) )

            else:
              if v in method_blks:
                # TODO Now I'm leaving incomplete dependency chain because I didn't close the circuit loop.
                # E.g. I do port.wr() somewhere in __main__ to write to a port.

                # Find total constraint (vb < blk) by vb=method_v < method_u=blk
                # INVALID if we have explicit constraint (blk < method_v) or (method_u < vb)

                v_blks = method_blks[ v ]
                for vb in v_blks:
                  if vb not in succ[u]:
                    for blk in assoc_blks:
                      if blk not in pred[v]:
                        if vb != blk:
                          if verbose: print("w<=0, v is method".center(10),v, blk)
                          if verbose: print(vb.__name__.center(25)," < ", \
                                      blk.__name__.center(25))
                          all_constraints.add( (vb, blk) )

              if (v, -1) not in visited:
                visited.add( (v, -1) )
                Q.append( (v, -1) ) # ? < v < u < ... < method < blk_id

        if w >= 0:
          for v in succ[u]:

            if v in all_upblks:
              # Find total constraint (blk < v) by blk=method_u' < method_u < v
              # INVALID if we have explicit constraint (method_u < blk)

              for blk in assoc_blks:
                if blk not in succ[u]:
                  if v != blk:
                    if verbose: print("w>=0, v is blk".center(10),blk, v)
                    if verbose: print(blk.__name__.center(25)," < ", \
                                      v.__name__.center(25))
                    all_constraints.add( (blk, v) )

            else:
              if v in method_blks:
                # assert v in method_blks, "Incomplete elaboration, something is wrong! %s" % hex(v)
                # TODO Now I'm leaving incomplete dependency chain because I didn't close the circuit loop.
                # E.g. I do port.wr() somewhere in __main__ to write to a port.

                # Find total constraint (blk < vb) by blk=method_u < method_v=vb
                # INVALID if we have explicit constraint (vb < method_u) or (method_v < blk)

                v_blks = method_blks[ v ]
                for vb in v_blks:
                  if not vb in pred[u]:
                    for blk in assoc_blks:
                      if not blk in succ[v]:
                        if vb != blk:
                          if verbose: print("w>=0, v is method".center(10), blk, v)
                          if verbose: print(blk.__name__.center(25)," < ", \
                                            vb.__name__.center(25))
                          all_constraints.add( (blk, vb) )

              if (v, 1) not in visited:
                visited.add( (v, 1) )
                Q.append( (v, 1) ) # blk_id < method < ... < u < v < ?


  def make_double_buffer_func( s ):

    strs = [ f"{repr(x)}._flip()" for x in s._dsl.all_signals if x._dsl.needs_double_buffer ]

    if not strs:
      def no_double_buffer():
        pass
      return no_double_buffer


    src = """
    def double_buffer():
      {}
    """.format( "\n      ".join(strs) )
    local = locals()
    exec(py.code.Source( src ).compile(), local)
    return local['double_buffer']

  # Construct the graph for update blocks

  vs  = all_upblks
  if isinstance( s, ComponentLevel2 ):
    vs -= all_update_ff

  es  = defaultdict(list)
  InD = { v:0 for v in vs }

  for (u, v) in list( all_constraints ): # u -> v, always
    InD[v] += 1
    es [u].append( v )

  # Perform topological sort for a serial schedule.

  serial_schedule = []
  Q = [ v for v in vs if not InD[v] ]
  while Q:
    random.shuffle(Q)
    #  print Q
    u = Q.pop()
    serial_schedule.append( u )
    for v in es[u]:
      InD[v] -= 1
      if not InD[v]:
        Q.append( v )

  if len(serial_schedule) != len(vs):
    raise UpblkCyclicError(
      'Update blocks have cyclic dependencies.'
      '* Please consult update dependency graph for details.'
    )


  if isinstance( s, ComponentLevel2 ):
    final_serial_schedule = [ make_double_buffer_func( s ) ]
    final_serial_schedule.extend( serial_schedule )
    final_serial_schedule.extend( all_update_ff )
  else:
    final_serial_schedule = serial_schedule

  assert final_serial_schedule, "No update block found in the model"

  if verbose:
    from graphviz import Digraph
    dot = Digraph()
    dot.graph_attr["rank"] = "same"
    dot.graph_attr["ratio"] = "compress"
    dot.graph_attr["margin"] = "0.1"

    for x in vs:
      dot.node( x.__name__, shape="box")

    for (x, y) in all_constraints:
      dot.edge( x.__name__,
                y.__name__ )

    dot.render( "/tmp/upblk-dag.gv", view=True )

  def tick_normal():
    for blk in final_serial_schedule:
      blk()
  s.tick = tick_normal
  s._dsl.schedule = final_serial_schedule

  # Clean up Signals

  def cleanup_signals( m ):
    if isinstance( m, list ):
      for i, o in enumerate( m ):
        if isinstance( o, Signal ):
          m[i] = o.default_value()
          m[i] <<= o.default_value()
        else:
          cleanup_signals( o )

    elif isinstance( m, NamedObject ):
      for name, obj in m.__dict__.items():
        if isinstance( name, str ) and name[0] != '_':
          if   isinstance( obj, Signal ):
            value = obj.default_value()
            value <<= obj.default_value()
            setattr( m, name, value )
          else:
            cleanup_signals( obj )

  cleanup_signals( s )

  def create_reset( top ):
    def reset():
      top.reset = Bits1(1)
      top.tick()
      top.tick()
      top.reset = Bits1(0)
    return reset

  s.sim_reset = create_reset( s )



"""
========================================================================
GetGiveIfc.py
========================================================================
RTL implementation of en/rdy micro-protocol.

Author: Yanghui Ou
  Date: Mar 19, 2019
"""
import greenlet

from pymtl3 import *
from pymtl3.dsl.errors import InvalidConnectionError
from pymtl3.stdlib.connects import connect_pairs

from .SendRecvIfc import RecvIfcRTL

#-------------------------------------------------------------------------
# GetIfcRTL
#-------------------------------------------------------------------------

class GetIfcRTL( CallerIfcRTL ):
  def construct( s, Type ):
    super().construct( en=True, rdy=True, MsgType=None, RetType=Type )

#-------------------------------------------------------------------------
# GiveIfcRTL
#-------------------------------------------------------------------------

class And( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort( Type )

    @s.update
    def up_and():
      s.out = s.in0 & s.in1

class GiveIfcRTL( CalleeIfcRTL ):
  def construct( s, Type ):
    super().construct( en=True, rdy=True, MsgType=None, RetType=Type )

  def connect( s, other, parent ):
    # We are doing GiveIfcRTL (s) -> [ AND ] -> RecvIfcRTL (other)
    # Basically we AND the rdy of both sides for enable
    if isinstance( other, RecvIfcRTL ):
      connect( s.ret, other.msg )

      m = And( Bits1 )

      if hasattr( parent, "give_recv_ander_cnt" ):
        cnt = parent.give_recv_ander_cnt
        setattr( parent, "give_recv_ander_" + str( cnt ), m )
      else:
        parent.give_recv_ander_cnt = 0
        parent.give_recv_ander_0   = m

      connect_pairs(
        m.in0, s.rdy,
        m.in1, other.rdy,
        m.out, s.en,
        m.out, other.en,
      )
      parent.give_recv_ander_cnt += 1
      return True

    elif isinstance( other, CalleeIfcCL ):
      if s._dsl.level <= other._dsl.level:
        raise InvalidConnectionError(
            "CL2RTL connection is not supported between RecvIfcRTL"
            " and CalleeIfcCL.\n"
            "          - level {}: {} (class {})\n"
            "          - level {}: {} (class {})".format(
                s._dsl.level, repr( s ), type( s ), other._dsl.level,
                repr( other ), type( other ) ) )

      m = GetRTL2GiveCL( s.MsgType )

      if hasattr( parent, "GetRTL2GiveCL_count" ):
        count = parent.GetRTL2GiveCL_count
        setattr( parent, "GetRTL2GiveCL_" + str( count ), m )
      else:
        parent.GetRTL2GiveCL_count = 0
        parent.GetRTL2GiveCL_0 = m

      connect_pairs(
        s,      m.get,
        m.give, other,
      )
      parent.GetRTL2GiveCL_count += 1
      return True

    return False

class GetIfcFL( CallerIfcFL ):

  def connect( s, other, parent ):

    # We are doing SendCL (other) -> [ RecvCL -> GiveIfcFL ] -> GetIfcFL (s)
    # SendCL is a caller interface
    if isinstance( other, CallerIfcCL ):
      m = RecvCL2GiveFL()

      if hasattr( parent, "RecvCL2GiveFL_count" ):
        count = parent.RecvCL2GiveFL_count
        setattr( parent, "RecvCL2GiveFL_" + str( count ), m )
      else:
        parent.RecvCL2GiveFL_count = 0
        parent.RecvCL2GiveFL_0 = m

      connect_pairs(
        other,  m.recv,
        m.give, s
      )
      parent.RecvCL2GiveFL_count += 1
      return True

    return False

class GiveIfcFL( CalleeIfcFL ):
  pass

#-------------------------------------------------------------------------
# GetRTL2GiveCL
#-------------------------------------------------------------------------

class GetRTL2GiveCL( Component ):

  def construct( s, MsgType ):
    # Interface
    s.get = GetIfcRTL( MsgType )

    s.entry = None

    @s.update
    def up_get_rtl():
      if s.entry is None and s.get.rdy:
        s.get.en = b1(1)
      else:
        s.get.en = b1(0)

    @s.update
    def up_entry():
      if s.get.en:
        s.entry = deepcopy( s.get.msg )

    s.add_constraints(
      U( up_get_rtl ) < M( s.give     ),
      U( up_get_rtl ) < M( s.give.rdy ),
      U( up_entry   ) < M( s.give     ),
      U( up_entry   ) < M( s.give.rdy ),
    )

  @non_blocking( lambda s : s.entry is not None )
  def give( s ):
    tmp = deepcopy( s.entry )
    s.entry = None
    return tmp

  def line_trace( s ):
    return "{}(){}".format( s.get, s.give )

#-------------------------------------------------------------------------
# RecvCL2SendRTL
#-------------------------------------------------------------------------

class RecvCL2GiveFL( Component ):

  @blocking
  def give( s ):
    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)
    ret = s.entry
    s.entry = None
    return ret

  @non_blocking( lambda s : s.entry is None )
  def recv( s, msg ):
    s.entry = msg

  def construct( s ):

    # Interface

    s.give = GiveIfcFL( method=s.give )

    s.entry = None

    s.add_constraints( M( s.recv ) > M( s.give ) ) # pipe behavior

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.give )



"""
========================================================================
HeuristicTopo.py
========================================================================

Author : Shunning Jiang
Date   : Dec 26, 2018
"""

import ast
from queue import PriorityQueue

from ..BasePass import BasePass, PassMetadata
from ..errors import PassOrderError
from ..sim.SimpleSchedulePass import SimpleSchedulePass, check_schedule

# FIXME also apply branchiness to all update_ff blocks

class CountBranches( ast.NodeVisitor ):

  def enter( self, node ):
    self.num_br = 0
    self.visit( node )
    return self.num_br

  def visit_If( self, node ):
    self.num_br += 1
    self.visit( node.test )

    for stmt in node.body:
      self.visit( stmt )

    for stmt in node.orelse:
      self.visit( stmt )

  def visit_IfExp( self, node ):
    self.num_br += 1
    self.visit( node.test )
    self.visit( node.body )
    self.visit( node.orelse )

  # A single for is estimated to be 10x of a branch
  def visit_For( self, node ):
    self.num_br += 0
    for stmt in node.body:
      self.visit( stmt )

  def visit_While( self, node ):
    self.num_br += 0
    for stmt in node.body:
      self.visit( stmt )

class HeuristicTopoPass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    self.schedule_intra_cycle( top )

    # Reuse simple's ff and flip schedule
    simple = SimpleSchedulePass()
    simple.schedule_ff( top )
    simple.schedule_posedge_flip( top )

  def schedule_intra_cycle( self, top ):

    # Construct the intra-cycle graph based on normal update blocks

    V   = top._dag.final_upblks - top.get_all_update_ff()
    E   = set()
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in top._dag.all_constraints: # u -> v
      if u in V and v in V:
        InD[v] += 1
        Es[u].append( v )
        E.add( (u, v) )

    # Extract branchiness

    # Initialize all generated net block to 0 branchiness
    branchiness = { x: 0 for x in top._dag.genblks }

    visitor = CountBranches()
    for blk in top.get_all_update_blocks():
      hostobj = top.get_update_block_host_component( blk )
      branchiness[ blk ] = visitor.enter( hostobj.get_update_block_info( blk )[-1] )

    # Perform topological sort for a serial schedule.
    # Note that here we use a priority queue to get the blocks with small
    # branchiness as early as possible

    top._sched.update_schedule = update_schedule = []

    # Python3 doesn't have hash for functions
    id_v = { id(v): v for v in V}

    Q = PriorityQueue(0)
    for v in V:
      if not InD[v]:
        Q.put( (branchiness[ v ], id(v)) )

    while not Q.empty():
      br, u = Q.get()
      update_schedule.append( id_v[u] )
      for v in Es[id_v[u]]:
        InD[v] -= 1
        if not InD[v]:
          Q.put( (branchiness[ v ], id(v)) )

    check_schedule( top, update_schedule, V, E, InD )



"""
========================================================================
Component.py
========================================================================
Add clk/reset signals.

Author : Yanghui Ou
  Date : Apr 6, 2019
"""
from collections import defaultdict

from pymtl3.datatypes import Bits1

from .ComponentLevel1 import ComponentLevel1
from .ComponentLevel7 import ComponentLevel7
from .Connectable import Const, InPort, Interface, MethodPort, OutPort, Signal, Wire
from .errors import InvalidAPICallError, InvalidConnectionError, NotElaboratedError
from .NamedObject import NamedObject
from .Placeholder import Placeholder


class Component( ComponentLevel7 ):

  # Override
  def _construct( s ):

    if not s._dsl.constructed:

      # clk and reset signals are added here.
      s.clk   = InPort( Bits1 )
      s.reset = InPort( Bits1 )

      # Merge the actual keyword args and those args set by set_parameter
      if s._dsl.param_tree is None:
        kwargs = s._dsl.kwargs
      elif s._dsl.param_tree.leaf is None:
        kwargs = s._dsl.kwargs
      else:
        kwargs = s._dsl.kwargs
        if "construct" in s._dsl.param_tree.leaf:
          more_args = s._dsl.param_tree.leaf[ "construct" ]
          kwargs.update( more_args )

      s._handle_decorated_methods()

      # Same as parent class _construct
      s.construct( *s._dsl.args, **kwargs )

      # We hook up the added clk and reset signals here. NOTE THAT if the
      # user overwrites clk/reset inside the component, we still get the
      # correct connection.
      parent = s.get_parent_object()
      if parent is not None:
        parent._connect_signal_signal( s.clk, parent.clk )
        parent._connect_signal_signal( s.reset, parent.reset )

      if s._dsl.call_kwargs is not None: # s.a = A()( b = s.b )
        s._continue_call_connect()

      s._dsl.constructed = True

  # This function deduplicates those checks in each API
  def _check_called_at_elaborate_top( s, func_name ):
    try:
      if s._dsl.elaborate_top is not s:
        raise InvalidAPICallError( func_name, s, s._dsl.elaborate_top )
    except AttributeError:
      raise NotElaboratedError()

  def _collect_objects_local( s, filt ):
    assert s._dsl.constructed
    ret = set()
    stack = []
    for (name, obj) in s.__dict__.items():
      if   isinstance( name, str ):
        if name[0] != '_': # filter private variables
          stack.append( obj )
    while stack:
      u = stack.pop()
      if filt( u ):
        ret.add( u )
      # ONLY LIST IS SUPPORTED
      elif isinstance( u, list ):
        stack.extend( u )
    return ret

  def _flush_pending_value_connections( s ):
    if s._dsl._has_pending_value_connections:
      s._dsl.all_value_nets = s._resolve_value_connections()
      s._dsl._has_pending_value_connections = False

  def _flush_pending_method_connections( s ):
    if s._dsl._has_pending_method_connections:
      s._dsl.all_method_nets = s._resolve_method_connections()
      s._dsl._has_pending_method_connections = False

  # These internal functions are implemented in a more generic way. The
  # public APIs should wrap around these functions.

  def _add_component( s, parent, name, indices, obj, provided_connections,
                      provided_upblk_reads, provided_upblk_writes, provided_upblk_calls,
                      provided_func_reads,  provided_func_writes, provided_func_calls ):
    try:
      top = s._dsl.elaborate_top
    except AttributeError:
      raise NotElaboratedError()

    top._dsl.elaborate_stack = [ parent ]

    # Check if we are adding obj to a list of to a component
    if not indices:
      # If we are adding field parent.x, we simply reuse the setattr hook
      assert not hasattr( parent, name ), f"Invalid add_component call: {parent} already has field {name}!"

      NamedObject.__setattr__ = NamedObject.__setattr_for_elaborate__
      setattr( parent, name, obj )
      del NamedObject.__setattr__

    else:
      assert hasattr( parent, name ), f"Invalid add_component call: Field '{name}' of {parent} should exist and be a list, but doesn't exist!"

      # We use the indices to get the list parent and set the element

      list_parent = getattr( parent, name )
      i = 0
      while i < len(indices) - 1:
        list_parent = list_parent[ indices[i] ]
        i += 1

      assert list_parent[ indices[i] ] is None, f"Invalid add_component call: {parent} already has field '{name}!'"
      list_parent[ indices[i] ] = obj

      # NOTE THAT we cannot use setattr hook for the top level metadata
      # because setting an element in the list doesn't call setattr.
      # As a result we have to copy some code from setattr hook and do it
      # manually here for the top level

      obj._dsl.parent_obj = parent
      obj._dsl.level      = parent._dsl.level + 1
      obj._dsl.my_name    = u_name = name + "".join( [ f"[{x}]" for x in indices ] )

      # Iterate through the param_tree and update u
      if parent._dsl.param_tree is not None:
        if parent._dsl.param_tree.children is not None:
          for comp_name, node in parent._dsl.param_tree.children.items():
            if comp_name == u_name:
              # Lazily create the param tree
              if obj._dsl.param_tree is None:
                obj._dsl.param_tree = ParamTreeNode()
              obj._dsl.param_tree.merge( node )

            elif node.compiled_re is not None:
              if node.compiled_re.match( u_name ):
                # Lazily create the param tree
                if obj._dsl.param_tree is None:
                  obj._dsl.param_tree = ParamTreeNode()
                obj._dsl.param_tree.merge( node )

      obj._dsl.full_name = ( parent._dsl.full_name + "." + u_name )

      # store the name/indices
      obj._dsl._my_name     = name
      obj._dsl._my_indices  = indices

      obj._dsl.elaborate_top = top
      top._dsl.elaborate_stack.append( obj )

      NamedObject.__setattr__ = NamedObject.__setattr_for_elaborate__
      obj._construct()
      del NamedObject.__setattr__

      top._dsl.elaborate_stack.pop()

    added_components = obj._collect_all_single( lambda x: isinstance( x, Component ) )

    # First elaborate all functions to spawn more named objects
    for c in added_components:
      c._elaborate_read_write_func()

    added_signals, added_method_ports = \
      obj._collect_all( [ lambda x: isinstance( x, Signal ), \
                          lambda x: isinstance( x, MethodPort ) ] )

    top._dsl.all_components    |= added_components
    top._dsl.all_signals       |= added_signals
    top._dsl.all_method_ports  |= added_method_ports

    top._dsl.all_named_objects |= added_components
    top._dsl.all_named_objects |= added_signals
    top._dsl.all_named_objects |= added_method_ports

    for c in added_components:
      top._collect_vars( c )

    # Lazy -- to avoid resolve_connection call which takes non-trivial
    # time upon adding any connect, I just mark pending here. Whenever you
    # call the right API which is get_all_value_nets()/get_method_nets(),
    # the API will batch flush all pending connections if the flag is
    # true. So please make sure to call the API whenever you want to get
    # the value/method nets.
    connection_pairs = []
    for (x, y) in provided_connections:
      connection_pairs.append( x )
      connection_pairs.append( eval(y) )
      if not top._dsl._has_pending_value_connections and isinstance( x, Signal ):
        top._dsl._has_pending_value_connections = True
      if not top._dsl._has_pending_method_connections and isinstance( x, MethodPort ):
        top._dsl._has_pending_method_connections = True

    # WE NEED TO ADD CONNECTIONS AT PARENT INSTEAD OF TOP
    parent.add_connections( *connection_pairs )

    # Now we put back the provided upblk metadata to parent and top
    for blk, obj_name in provided_upblk_reads:
      parent._dsl.upblk_reads[blk].add( eval(obj_name) )

    for blk, obj_name in provided_upblk_writes:
      parent._dsl.upblk_writes[blk].add( eval(obj_name) )

    for blk, obj_name in provided_upblk_calls:
      parent._dsl.upblk_calls[blk].add( eval(obj_name) )

    for func, obj_name in provided_func_reads:
      parent._dsl.func_reads[func].add( eval(obj_name) )

    for func, obj_name in provided_func_writes:
      parent._dsl.func_writes[func].add( eval(obj_name) )

    for func, obj_name in provided_func_calls:
      parent._dsl.func_calls[func].add( eval(obj_name) )

    del top._dsl.elaborate_stack

  def _delete_component( top, obj ):

    # This nested delete function is to create an extra layer to properly
    # call garbage collector. If we can make sure it is collected
    # automatically and fast enough, we might remove this extra layer
    #
    # EDIT: After experimented w/ and w/o gc.collect(), it seems like it
    # is eventually collected, but sometimes the intermediate memory
    # footprint might reach up to gigabytes, so let's keep the
    # gc.collect() for now

    def _delete_component_internal( top, foo ):
      assert obj._dsl.elaborate_top is top

      # First make sure we flush pending connections
      value_nets  = top.get_all_value_nets()
      method_nets = top.get_all_method_nets()

      # Delete the placeholder from parent. We need to see if we are
      # dealing with a list parent or a component parent.

      parent = foo.get_parent_object()

      if foo._dsl._my_indices:
        # We use the saved indices to get the list parent and set the index
        my_indices = foo._dsl._my_indices
        list_parent = getattr( parent, foo._dsl._my_name )
        i = 0
        while i < len(my_indices) - 1:
          list_parent = list_parent[ my_indices[i] ]
          i += 1
        list_parent[ my_indices[i] ] = None
      else:
        # Simply delete the attribute if it's merely a field
        delattr( parent, foo._dsl.my_name )

      # Remove all components, signals, and method ports
      removed_components, removed_signals, removed_method_ports = \
        foo._collect_all( [ lambda x: isinstance( x, Component ), \
                            lambda x: isinstance( x, Signal ), \
                            lambda x: isinstance( x, MethodPort ) ] )

      top._dsl.all_components    -= removed_components
      top._dsl.all_signals       -= removed_signals
      top._dsl.all_method_ports  -= removed_method_ports

      top._dsl.all_named_objects -= removed_components

      removed_connectables = removed_signals | removed_method_ports
      top._dsl.all_named_objects -= removed_connectables

      removed_consts = set()
      if isinstance( foo, Placeholder ):
        # No need to uncollect vars from a placeholder
        assert not foo._dsl.consts
      else:
        for x in removed_components:
          # remove consts
          removed_consts |= x._dsl.consts
          # uncollect variables
          top._uncollect_vars( x )

      saved_upblk_reads  = []
      saved_upblk_writes = []
      saved_upblk_calls  = []
      saved_func_reads   = []
      saved_func_writes  = []
      saved_func_calls   = []

      # If an update block/function in parent writes the inport or reads
      # the outport or calls a callee method of the deleted component, we
      # must save the information (upA reads B) to avoid bugs or
      # explicitly re-elaborating the parent.

      for blk, reads in parent._dsl.upblk_reads.items():
        assert blk in top._dsl.all_upblk_reads
        to_save = set()
        for x in reads:
          if x in removed_connectables:
            to_save.add( x )
            saved_upblk_reads.append( (blk, repr(x)) )
        parent._dsl.upblk_reads[blk] -= to_save

      for blk, writes in parent._dsl.upblk_writes.items():
        assert blk in top._dsl.all_upblk_writes
        to_save = set()
        for x in writes:
          if x in removed_connectables:
            to_save.add( x )
            saved_upblk_writes.append( (blk, repr(x)) )
        parent._dsl.upblk_writes[blk] -= to_save

      for blk, calls in parent._dsl.upblk_calls.items():
        assert blk in top._dsl.all_upblk_calls
        to_save = set()
        for x in calls:
          if x in removed_connectables:
            to_save.add( x )
            saved_upblk_calls.append( (blk, repr(x)) )
        parent._dsl.upblk_calls[blk] -= to_save

      # We need to save the information for funcs too
      for func, reads in parent._dsl.func_reads.items():
        to_save = set()
        for x in reads:
          if x in removed_connectables:
            to_save.add( x )
            saved_func_reads.append( (func, repr(x)) )
        parent._dsl.func_reads[func] -= to_save

      for func, writes in parent._dsl.func_writes.items():
        to_save = set()
        for x in writes:
          if x in removed_connectables:
            to_save.add( x )
            saved_func_writes.append( (func, repr(x)) )
        parent._dsl.func_writes[func] -= to_save

      for func, calls in parent._dsl.func_calls.items():
        to_save = set()
        for x in calls:
          if x in removed_connectables:
            to_save.add( x )
            saved_func_calls.append( (func, repr(x)) )
        parent._dsl.func_calls[func] -= to_save

      saved_connections = []

      for x in removed_connectables:
        # Clean up all_adjancency at top
        if x in top._dsl.all_adjacency:
          for other in top._dsl.all_adjacency[x]:
            # other must be in the dict
            # If other will be removed, we don't need to remove it here ..
            if other not in removed_connectables and other not in removed_consts:
              top._dsl.all_adjacency[other].remove( x )
              if isinstance( other, Const ):
                other = other._dsl.const
              saved_connections.append( (other, "top"+repr(x)[1:]) ) # other is from outside
          del top._dsl.all_adjacency[x]

        # Clean up adjacency at parent
        if x in parent._dsl.adjacency:
          for other in parent._dsl.adjacency[x]:
            # other must be in the dict
            if other not in removed_connectables:
              parent._dsl.adjacency[other].remove( x )
          del parent._dsl.adjacency[x]

      for x in removed_components:
        del x._dsl.parent_obj
        del x._dsl.elaborate_top
      for x in removed_connectables:
        del x._dsl.parent_obj
        del x._dsl.elaborate_top
        x._dsl.full_name = "<deleted>"+x._dsl.full_name
      for y in removed_consts:
        del y._dsl.parent_obj

      # We don't break nets anymore. Instead, we set the flags to true so
      # that the next get_xxx_net will immediately recollect nets.
      top._dsl._has_pending_value_connections = True
      top._dsl._has_pending_method_connections = True

      # We clean up the connect_order list. If we want to preserve the
      # original connect order, we can play some other tricks here such as
      # leaving the strings in there. However I don't think that's clean
      # enough. Thus I'm just removing them right now.
      new_connect_order = []
      for (x, y) in parent._dsl.connect_order:
        if x not in removed_signals and y not in removed_signals: # TODO method port
          new_connect_order.append( (x, y) )

      parent._dsl.connect_order = new_connect_order

      return saved_connections, saved_upblk_reads, saved_upblk_writes, saved_upblk_calls, \
             saved_func_reads, saved_func_writes, saved_func_calls

    return _delete_component_internal( top, obj )
    # import gc
    # gc.collect() # this takes 0.1 seconds

  #-----------------------------------------------------------------------
  # Post-elaborate public APIs (can only be called after elaboration)
  #-----------------------------------------------------------------------

  """ Convenience/utility APIs """

  def apply( s, pass_instance ):
    try:
      import pypyjit
      pypyjit.set_param("off")
    except:
      pass

    assert type(pass_instance) is not type, f"Should pass in a pass instance like " \
                                            f"'{pass_instance.__name__}()' instead of '{pass_instance.__name__}'"
    assert callable( pass_instance ), f"Should override __call__ of {pass_instance.__name__} for a valid pass"
    pass_instance( s )

    try:
      pypyjit.set_param("default")
      pypyjit.set_param("trace_limit=100000")
    except:
      pass

  def check( s ):
    s._check_valid_dsl_code()

  # TODO maybe we should implement these two lock/unlock APIs as passes?
  # They expose kernel implementation details though ...
  def lock_in_simulation( s ):
    s._check_called_at_elaborate_top( "lock_in_simulation" )

    swapped_signals = defaultdict(list)

    # Swap all Signal objects with actual data

    Q = [ (s, s) ]
    while Q:
      current_obj, host = Q.pop()
      if isinstance( current_obj, list ):
        for i, obj in enumerate( current_obj ):
          if isinstance( obj, Signal ):
            try:
              current_obj[i] = obj.default_value()
            except Exception as e:
              raise type(e)(str(e) + f' happens at {obj!r}')

            try:
              current_obj[i] <<= obj.default_value()
            except Exception:
              pass

            swapped_signals[ host ].append( (current_obj, i, obj, True) )

          elif isinstance( obj, Component ):
            Q.append( (obj, obj) )
          elif isinstance( obj, (Interface, list) ):
            Q.append( (obj, host) )

      elif isinstance( current_obj, NamedObject ):
        for i, obj in current_obj.__dict__.items():
          if i[0] != '_': # impossible to have tuple
            if isinstance( obj, Signal ):
              try:
                value = obj.default_value()
              except Exception as e:
                raise type(e)(str(e) + f' happens at {obj!r}')

              try:
                value <<= obj.default_value()
              except Exception:
                pass

              setattr( current_obj, i, value )

              swapped_signals[ host ].append( (current_obj, i, obj, False) )

            elif isinstance( obj, Component ):
              Q.append( (obj, obj) )
            elif isinstance( obj, (Interface, list) ):
              Q.append( (obj, host) )

    s._dsl.swapped_signals = swapped_signals
    s._dsl.locked_simulation = True

  def unlock_simulation( s ):
    s._check_called_at_elaborate_top( "unlock_simulation" )
    try:
      assert s._dsl.locked_simulation
    except:
      raise AttributeError("Cannot unlock an unlocked/never locked model.")

    swapped_values  = defaultdict(list)
    for component, records in s._dsl.swapped_signals.items():
      for current_obj, i, obj, is_list in records:
        if is_list:
          swapped_values[ component ].append( (current_obj, i, current_obj[i], is_list) )
          current_obj[i] = obj
        else:
          swapped_values[ component ].append( (current_obj, i, getattr(current_obj, i), is_list) )
          setattr( current_obj, i, obj )

    s._dsl.swapped_values = swapped_values
    s._dsl.locked_simulation = False

  """ APIs that provide local metadata of a component """

  def get_component_level( s ):
    try:
      return s._dsl.level
    except AttributeError:
      raise NotElaboratedError()

  def get_child_components( s ):
    return s._collect_objects_local( lambda x: isinstance( x, Component ) )

  def get_input_value_ports( s ):
    return s._collect_objects_local( lambda x: isinstance( x, InPort ) )

  def get_output_value_ports( s ):
    return s._collect_objects_local( lambda x: isinstance( x, OutPort ) )

  def get_wires( s ):
    return s._collect_objects_local( lambda x: isinstance( x, Wire ) )

  def get_update_blocks( s ):
    assert s._dsl.constructed
    return s._dsl.upblks

  def get_update_ff( s ):
    assert s._dsl.constructed
    return s._dsl.update_ff

  def get_upblk_metadata( s ):
    assert s._dsl.constructed
    return s._dsl.upblk_reads, s._dsl.upblk_writes, s._dsl.upblk_calls

  # These xxx_order APIs should be used when some pass wants to process
  # connections/upblks in the order that they are created

  def get_connect_order( s ):
    try:
      return s._dsl.connect_order
    except AttributeError:
      raise NotElaboratedError()

  def get_update_block_order( s ):
    try:
      return s._dsl.upblk_order
    except AttributeError:
      raise NotElaboratedError()

  """ Metadata APIs that should only be called at elaborated top"""

  def get_all_object_filter( s, filt ):
    assert callable( filt )
    try:
      return { x for x in s._dsl.all_named_objects if filt(x) }
    except AttributeError:
      return s._collect_all_single( filt )

  def get_local_object_filter( s, filt ):
    assert callable( filt )
    return s._collect_objects_local( filt )

  def get_all_components( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_blocks" )
      return s._dsl.all_components
    except AttributeError:
      raise NotElaboratedError()

  def get_all_update_blocks( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_blocks" )
      return s._dsl.all_upblks
    except AttributeError:
      raise NotElaboratedError()

  def get_all_update_ff( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_ff" )
      return s._dsl.all_update_ff
    except AttributeError:
      raise NotElaboratedError()

  def get_all_upblk_metadata( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_upblk_metadata" )
      return s._dsl.all_upblk_reads, s._dsl.all_upblk_writes, s._dsl.all_upblk_calls
    except AttributeError:
      raise NotElaboratedError()

  def get_all_explicit_constraints( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_explicit_constraints" )
      return s._dsl.all_U_U_constraints, s._dsl.all_RD_U_constraints, \
             s._dsl.all_WR_U_constraints, s._dsl.all_M_constraints
    except AttributeError:
      raise NotElaboratedError()

  # is_lambda?, src, line, filename, ast
  def get_update_block_info( s, blk ):
    try:
      name_info = s.__class__._name_info
    except AttributeError: # This component doesn't have update block
      return None

    return name_info[blk.__name__]

  def get_update_block_host_component( s, blk ):
    try:
      s._check_called_at_elaborate_top( "get_update_block_host_component" )
      return s._dsl.all_upblk_hostobj[ blk ]
    except AttributeError:
      raise NotElaboratedError()

  def get_all_method_nets( s ):
    s._check_called_at_elaborate_top( "get_all_method_nets" )
    s._flush_pending_method_connections()
    return s._dsl.all_method_nets

  # Override
  def get_all_value_nets( s ):
    s._check_called_at_elaborate_top( "get_all_value_nets" )
    s._flush_pending_value_connections()
    return s._dsl.all_value_nets

  def get_signal_adjacency_dict( s ):
    try:
      s._check_called_at_elaborate_top( "get_signal_adjacency_dict" )
    except AttributeError:
      raise NotElaboratedError()
    return s._dsl.all_adjacency

  """ Mutation APIs to add/delete components and connections"""

  # Shunning: These API implement replacing a component with another
  # module. This is the first mutation API we want to test thoroughly

  # TODO test everything below and figure out whether we need to delete
  # a normal component

  def replace_component( top, foo, cls, check=True ):
    top._check_called_at_elaborate_top( "replace_component" )

    parent = foo.get_parent_object()
    foo_name    = foo._dsl._my_name
    foo_indices = foo._dsl._my_indices

    saved_connections, saved_upblk_reads, saved_upblk_writes, saved_upblk_calls, \
      saved_func_reads, saved_func_writes, saved_func_calls = top._delete_component( foo )

    new_obj = cls( *foo._dsl.args, **foo._dsl.kwargs )

    # We actually don't need to merge param tree here because when we call
    # _add_component, the parameters stored in parent will be pushed down
    # to new_obj
    top._add_component( parent, foo_name, foo_indices, new_obj, saved_connections,
                        saved_upblk_reads, saved_upblk_writes, saved_upblk_calls,
                        saved_func_reads, saved_func_writes, saved_func_calls)

    top._flush_pending_value_connections()
    top._flush_pending_method_connections()
    if check:
      top.check()

  def replace_component_with_obj( top, foo, new_obj, check=True ):
    top._check_called_at_elaborate_top( "replace_component" )

    parent = foo.get_parent_object()
    foo_name    = foo._dsl._my_name
    foo_indices = foo._dsl._my_indices

    saved_connections, saved_upblk_reads, saved_upblk_writes, saved_upblk_calls, \
      saved_func_reads, saved_func_writes, saved_func_calls = top._delete_component( foo )

    # We actually don't need to merge param tree here because when we call
    # _add_component, the parameters stored in parent will be pushed down
    # to new_obj
    top._add_component( parent, foo_name, foo_indices, new_obj, saved_connections,
                        saved_upblk_reads, saved_upblk_writes, saved_upblk_calls,
                        saved_func_reads, saved_func_writes, saved_func_calls)

    top._flush_pending_value_connections()
    top._flush_pending_method_connections()
    if check:
      top.check()

  def add_connection( s, o1, o2 ):
    try:
      top = s._dsl.elaborate_top
    except AttributeError:
      raise NotElaboratedError()

    try:
      s._connect( o1, o2, internal=False )
    except AssertionError as e:
      raise InvalidConnectionError( "\n{}".format(e) )

    for x, adjs in s._dsl.adjacency.items():
      top._dsl.all_adjacency[x].update( adjs )

  def add_connections( s, *args ):
    try:
      top = s._dsl.elaborate_top
    except AttributeError:
      raise NotElaboratedError()

    if len(args) & 1 != 0:
       raise InvalidConnectionError( "Odd number ({}) of objects provided.".format( len(args) ) )

    for i in range(len(args)>>1):
      try:
        s._connect( args[ i<<1 ], args[ (i<<1)+1 ], internal=False )
      except InvalidConnectionError as e:
        raise InvalidConnectionError( "\n- In connect_pair, when connecting {}-th argument to {}-th argument\n{}\n " \
              .format( (i<<1)+1, (i<<1)+2 , e ) )

    for x, adjs in s._dsl.adjacency.items():
      top._dsl.all_adjacency[x].update( adjs )

  # TODO implement everything below and test them

  # Override
  def add_component_by_name( s, name, obj, provided_connections=[] ):
    raise NotImplementedError()

  # Override
  def delete_component( s, name ):
    raise NotImplementedError()

  # Do we still need disconnect and disconnect_pair?



#=========================================================================
# StructuralTranslatorL3.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 4, 2019
"""Provide L3 structural translator."""

from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir import StructuralRTLIRSignalExpr as sexp
from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL3Pass import (
    StructuralRTLIRGenL3Pass,
)

from .StructuralTranslatorL2 import StructuralTranslatorL2


class StructuralTranslatorL3( StructuralTranslatorL2 ):

  #-----------------------------------------------------------------------
  # gen_structural_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def gen_structural_trans_metadata( s, tr_top ):
    tr_top.apply( StructuralRTLIRGenL3Pass( s.inst_conns ) )

  #-----------------------------------------------------------------------
  # translate_structural
  #-----------------------------------------------------------------------

  # Override
  def translate_structural( s, tr_top ):
    s.structural.decl_ifcs = {}
    super().translate_structural( tr_top )

  #-----------------------------------------------------------------------
  # translate_decls
  #-----------------------------------------------------------------------

  # Override
  def translate_decls( s, m ):
    m_rtype = m._pass_structural_rtlir_gen.rtlir_type

    # Interfaces
    ifc_decls = []
    for ifc_id, rtype in m_rtype.get_ifc_views_packed():
      if isinstance( rtype, rt.Array ):
        array_rtype = rtype
        ifc_rtype = rtype.get_sub_type()
      else:
        array_rtype = None
        ifc_rtype = rtype

      # Translate all ports and interfaces inside the interface
      ports = []
      all_properties = ifc_rtype.get_all_properties_packed()
      for p_id, _p_rtype in all_properties:
        if isinstance( _p_rtype, rt.Array ):
          p_array_rtype = _p_rtype
          p_rtype = _p_rtype.get_sub_type()
        else:
          p_array_rtype = None
          p_rtype = _p_rtype

        ports.append( s.rtlir_tr_interface_port_decl(
          m,
          s.rtlir_tr_var_id( p_id ),
          p_rtype,
          s.rtlir_tr_unpacked_array_type( p_array_rtype ),
        ) )
      ifc_decls.append(
        s.rtlir_tr_interface_decl(
          ifc_id,
          ifc_rtype,
          s.rtlir_tr_unpacked_array_type( array_rtype ),
          s.rtlir_tr_interface_port_decls( ports )
      ) )
    s.structural.decl_ifcs[m] = s.rtlir_tr_interface_decls( ifc_decls )

    super().translate_decls( m )

  #-----------------------------------------------------------------------
  # rtlir_signal_expr_translation
  #-----------------------------------------------------------------------

  # Override
  def rtlir_signal_expr_translation( s, expr, m ):
    """Translate a signal expression in RTLIR into its backend representation.

    Add support for the following operations at L3: sexp.InterfaceAttr
    """
    if isinstance( expr, sexp.InterfaceAttr ):
      return s.rtlir_tr_interface_attr(
        s.rtlir_signal_expr_translation(expr.get_base(), m), expr.get_attr())

    elif isinstance( expr, sexp.InterfaceViewIndex ):
      return s.rtlir_tr_interface_array_index(
        s.rtlir_signal_expr_translation(expr.get_base(), m), expr.get_index())

    else:
      return super().rtlir_signal_expr_translation(expr, m)

  #-----------------------------------------------------------------------
  # Methods to be implemented by the backend translator
  #-----------------------------------------------------------------------

  # Declarations
  def rtlir_tr_interface_port_decls( s, port_decls ):
    raise NotImplementedError()

  def rtlir_tr_interface_port_decl( s, m, port_id, port_rtype, port_array_type ):
    raise NotImplementedError()

  def rtlir_tr_interface_decls( s, ifc_decls ):
    raise NotImplementedError()

  def rtlir_tr_interface_decl( s, ifc_id, ifc_rtype, array_type, port_decls ):
    raise NotImplementedError()

  # Signal operations
  def rtlir_tr_interface_array_index( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_interface_attr( s, base_signal, attr ):
    raise NotImplementedError()



"""
========================================================================
Interface_test.py
========================================================================

Author : Shunning Jiang, Yanghui Ou
Date   : Jan 1, 2018
"""
from collections import deque

from pymtl3.datatypes import Bits1, Bits8, Bits128
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3, connect
from pymtl3.dsl.Connectable import InPort, Interface, OutPort, Wire

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  print()
  T, time = 0, 20
  while not A.done() and T < time:
    A.tick()
    print(A.line_trace())
    T += 1

def valrdy_to_str( msg, val, rdy ):

  str_   = f"{msg}"
  nchars = max( len( str_ ), 15 )

  if       val and not rdy:
    str_ = "#".ljust( nchars )
  elif not val and     rdy:
    str_ = " ".ljust( nchars )
  elif not val and not rdy:
    str_ = ".".ljust( nchars )

  return str_.ljust( nchars )

class InValRdyIfc( Interface ):

  def construct( s, Type ):

    s.msg = InPort( Type )
    s.val = InPort( int if Type is int else Bits1 )
    s.rdy = OutPort( int if Type is int else Bits1 )

  def line_trace( s ):
    return valrdy_to_str( s.msg, s.val, s.rdy )

class OutValRdyIfc( Interface ):

  def construct( s, Type ):

    s.msg = OutPort( Type )
    s.val = OutPort( int if Type is int else Bits1 )
    s.rdy = InPort( int if Type is int else Bits1 )

  def line_trace( s ):
    return valrdy_to_str( s.msg, s.val, s.rdy )

class TestSourceValRdy( ComponentLevel3 ):

  def construct( s, Type, msgs ):
    assert type(msgs) == list, "TestSrc only accepts a list of inputs!"

    s.msgs    = msgs
    s.src_msgs = deque( msgs )
    s.default = Type()
    s.out     = OutValRdyIfc( Type )

    @s.update_ff
    def up_src():
      if (s.out.rdy & s.out.val) and s.src_msgs:
        s.src_msgs.popleft()
      s.out.val <<= Bits1( len(s.src_msgs) > 0 )
      s.out.msg <<= s.default if not s.src_msgs else Type(s.src_msgs[0])

  def done( s ):
    return not s.src_msgs

  def line_trace( s ):
    return s.out.line_trace()

class TestSinkValRdy( ComponentLevel3 ):

  def construct( s, Type, msgs ):
    assert type(msgs) == list, "TestSink only accepts a list of outputs!"

    s.msgs = msgs
    s.sink_msgs = deque( s.msgs )

    s.in_ = InValRdyIfc( Type )

    @s.update_ff
    def up_sink():
      s.in_.rdy <<= Bits1( len(s.sink_msgs) > 0 )

      if s.in_.val and s.in_.rdy:
        ref = s.sink_msgs.popleft()
        ans = s.in_.msg

        assert ref == ans, "Expect {}, get {} instead".format(ref, ans)

  def done( s ):
    return not s.sink_msgs

  def line_trace( s ):
    return s.in_.line_trace()

def test_simple():

  class Top( ComponentLevel3 ):

    def construct( s ):

      s.src  = TestSourceValRdy( Bits8, [ 0, 1, 2 ] )
      s.sink = TestSinkValRdy  ( Bits8, [ 0, 1, 2 ] )( in_ = s.src.out )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_nested_port_bundle():

  class ValRdyBundle( Interface ):

    def construct( s, Type=int ):
      s.msg = Wire( Type )
      s.val = Wire( Bits1 )
      s.rdy = Wire( Bits1 )

    def line_trace( s ):
      return valrdy_to_str( s.msg, s.val, s.rdy )

  class SuperBundle( Interface ):

    def construct( s, Type ):
      s.req  = [ [ ValRdyBundle(Type) for i in range(4) ] for j in range(4) ]

  class Top( ComponentLevel3 ):

    def construct( s ):

      s.src  = [ TestSourceValRdy( Bits8, [ i,i+1,i+2,i,i+1,i+2 ] ) for i in range(4) ]
      # (0+1+2+3)*4=24, (1+2+3+4)*4=40, (2+3+4+5)*5=56
      s.sink = TestSinkValRdy  ( Bits8, [ 24, 40, 56, 24, 40, 56] )

      s.sb = SuperBundle( Bits8 )
      s.wire = [ [ Wire(Bits8) for i in range(4) ] for j in range(4) ]

      for i in range(4):
        connect( s.src[i].out.rdy, s.sink.in_.rdy )
        for j in range(4):
          connect( s.sb.req[i][j].msg, s.src[i].out.msg )
          connect( s.wire[i][j],       s.sb.req[i][j].msg )

      @s.update
      def up_from_req():
        s.sink.in_.val = 1
        s.sink.in_.msg = 0
        for i in range(4):
          for j in range(4):
            s.sink.in_.msg += s.wire[i][j]

    def done( s ):
      return s.sink.done() and any( src.done() for src in s.src )

    def line_trace( s ):
      return "|".join( [ x.line_trace() for x in s.src] ) + " >>> " + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_customized_connect():

  class MockRecvIfc( Interface ):
    def construct( s ):
      s.recv_msg = InPort( Bits1 )
      s.recv_val = InPort( Bits1 )

    def connect( s, other, parent ):
      if isinstance( other, MockSendIfc ):
        s.recv_msg //= other.send_msg
        s.recv_val //= other.send_val
        return True

      return False

  class MockSendIfc( Interface ):
    def construct( s ):
      s.send_msg = OutPort( Bits1 )
      s.send_val = OutPort( Bits1 )

    def connect( s, other, parent ):
      if isinstance( other, MockRecvIfc ):
        s.send_msg //= other.recv_msg
        s.send_val //= other.recv_val
        return True

      return False

  class A( ComponentLevel3 ):
    def construct( s ):
      s.send = MockSendIfc()

      @s.update
      def up_send():
        s.send.send_msg = Bits1( 1 )
        s.send.send_val = Bits1( 1 )

  class B( ComponentLevel3 ):
    def construct( s ):
      s.recv = MockRecvIfc()

      @s.update
      def up_recv():
        print("recv_msg", s.recv.recv_msg, "recv_val", s.recv.recv_val)

  class Top( ComponentLevel3 ):

    def construct( s ):
      s.a = A()
      s.b = B()
      connect( s.a.send, s.b.recv )

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  _test_model( Top )

#-------------------------------------------------------------------------
# Test customized connect with adapters
#-------------------------------------------------------------------------

def test_customized_connect_adapter():

  class MockAdapter( ComponentLevel3 ):

    count = 0

    def construct( s, InType, OutType ):

      s.in_ = InPort ( InType  )
      s.out = OutPort( OutType )

      @s.update
      def adapter_incr():
        s.out = s.in_ + OutType( 1 )

  class MockRecvIfc( Interface ):

    def construct( s, Type ):
      s.recv_msg = InPort( Type  )
      s.recv_val = InPort( Bits1 )

      s.Type = Type

    def connect( s, other, parent ):
      if isinstance( other, MockSendIfc ):

        MockAdapter.count += 1
        m = MockAdapter( other.Type, s.Type )
        setattr( parent, "mock_adapter_" + str( MockAdapter.count ), m )

        other.send_msg //= m.in_
        m.out          //= s.recv_msg
        other.send_val //= s.recv_val
        return True

      return False

  class MockSendIfc( Interface ):
    def construct( s, Type ):
      s.send_msg = OutPort( Type  )
      s.send_val = OutPort( Bits1 )

      s.Type = Type

    def connect( s, other, parent ):
      if isinstance( other, MockRecvIfc ):
        s.send_msg //= other.recv_msg
        s.send_val //= other.recv_val
        return True

      return False

  class A( ComponentLevel3 ):
    def construct( s ):
      s.send = [ MockSendIfc( Bits8 ) for _ in range( 10 ) ]

      @s.update
      def up_send():
        for i in range( 10 ):
          s.send[i].send_msg = Bits1( 1 )
          s.send[i].send_val = Bits1( 1 )

  class B( ComponentLevel3 ):
    def construct( s ):
      s.recv = [ MockRecvIfc( Bits128 ) for _ in range( 10 ) ]

      @s.update
      def up_recv():
        for i in range( 10 ):
          print("recv_msg", i, s.recv[i].recv_msg, "recv_val", s.recv[i].recv_val)

  class Top( ComponentLevel3 ):

    def construct( s ):
      s.a = A()
      s.b = B()
      for i in range( 10 ):
        connect( s.b.recv[i], s.a.send[i] )

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  _test_model( Top )
